#!/usr/bin/env python

import  MySQLdb as mysqldb
from    datetime import datetime

class DBConfig(object):
    """@brief responsible for holding the attributes if the database configuration."""

    DEFAULT_TCP_PORT = 3306

    def __init__(self):
        self.serverAddress      =   ""
        self.serverPort         =   DBConfig.DEFAULT_TCP_PORT
        self.username           =   ""
        self.password           =   ""
        self.dataBaseName       =   ""
        self.uio                =   None

class DatabaseIFError(Exception):
  pass

class DatabaseIF(object):
    """@brief Responsible for providing an interface to a database to allow
              Creation of database
              Creation of tables
              Execution of database sql commands
              """

    """@brief Set this to True to see the SQL statements executed."""
    DEBUG=False

    @staticmethod
    def CheckTableExists(connection, tableName):
        """@brief Check if a table exists in the selected database.
           @param connection The connection to the database.
           @param tableName The name of the table to check for.
           @return True if the table exists, False if not."""
        cursor = connection.cursor()
        cmd="""SELECT COUNT(*) FROM information_schema.tables WHERE table_name = '{0}'""".format(tableName.replace('\'', '\'\''))
        cursor.execute(cmd)
        tableExists = cursor.fetchone()[0] == 1
        cursor.close()
        return tableExists

    @staticmethod
    def GetValidColName(colName):
        """@brief Get a valid database column name."""

        VALID_CHAR_LIST = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ$_"
        for aChar in colName:
            if aChar not in VALID_CHAR_LIST:
                colName = colName.replace(aChar, '_')

        return colName

    @staticmethod
    def CreateTable(connection, tableName, tableSchemaDict):
        """"@brief Create a table in the currently USED database..
            @param connection The connection to the database.
            @param param:
            @param tableSchemaDict A python dictionary that defines the table schema.
                                   Each dictionary key is the name of the column in the table.
                                   Each associated value is the SQL definition of the column type (E.G VARCHAR(64), FLOAT(5,2) etc)."""
        cursor = connection.cursor()

        sqlCmd = 'CREATE TABLE IF NOT EXISTS `{}` ('.format(tableName)
        for colName in list(tableSchemaDict.keys()):
            colDef = tableSchemaDict[colName]
            correctedColName = DatabaseIF.GetValidColName(colName)
            sqlCmd = sqlCmd + "`{}` {},\n".format(correctedColName, colDef)

        sqlCmd = sqlCmd[:-2]
        sqlCmd = sqlCmd + ");"
        DatabaseIF.ExecuteSQL(connection, cursor, sqlCmd)
        cursor.close()

    @staticmethod
    def GetQuotedValue(value):
        return '\"{}"'.format(str(value))

    @staticmethod
    def InsertRow(connection, tableName, myDict):
        """@brief Insert a row into the table."""
        cursor = connection.cursor()
        timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        myDict['TIMESTAMP'] = timestamp

        keyList = list(myDict.keys())
        valueList = []
        for key in keyList:
            valueList.append(str(myDict[key]))

        sql = 'INSERT INTO `' + tableName
        sql += '` ('
        sql += ', '.join(keyList)
        sql += ') VALUES ('
        sql += ', '.join(map(DatabaseIF.GetQuotedValue, valueList))
        sql += ');'
        DatabaseIF.ExecuteSQL(connection, cursor, sql)
        cursor.close()

    @staticmethod
    def ExecuteSQL(connection, cursor, cmd):
        try:
            if DatabaseIF.DEBUG:
                print('ExecuteSQL(): ',cmd)
            rowsAffected = cursor.execute(cmd)
            connection.commit()
        except:
            connection.rollback();
            raise

        return rowsAffected

    @staticmethod
    def GetInsertableDict(self, rxDict, tableSchema):
        """@brief Convert the dict into a dict that can be stored in the database.
                  All the keys must have names that only contain characters that are valid for table columns.
                  Only those keys that are in the tableSchema will be inserted."""
        keys = list(rxDict.keys())
        convertedDict = {}
        for key in keys:
            colName = DatabaseIF.GetValidColName(key)
            #Only add the keys that we wish to store in the database
            if colName in list(tableSchema.keys()):
                convertedDict[colName] = rxDict[key]
        return convertedDict

    def __init__(self, config):
        """@brief Constructor
           @param config The database configuration instance."""
        self._dbConfig = config

        self._dbCon = None

    def _info(self, msg):
        if self._dbConfig.uio:
            self._dbConfig.uio.info(msg)


    def _debug(self, msg):
        if self._dbConfig.uio.debug:
            self._dbConfig.uio.debug(msg)
            
    def connect(self):
        """@brief connect to the database server."""
        self._info("Connecting to {}:{} (database = {})".format(self._dbConfig.serverAddress, self._dbConfig.serverPort, self._dbConfig.dataBaseName))
        self._dbCon = mysqldb.Connection(host=self._dbConfig.serverAddress,\
                                                port=self._dbConfig.serverPort,\
                                                user=self._dbConfig.username,\
                                                passwd=self._dbConfig.password,\
                                                db=self._dbConfig.dataBaseName)
        self._info("Connected")

    def connectNoDB(self):
        """@brief connect to the database server."""
        self._info("Connecting to {}:{}".format(self._dbConfig.serverAddress, self._dbConfig.serverPort))
        self._dbCon = mysqldb.Connection(host=self._dbConfig.serverAddress,\
                                                port=self._dbConfig.serverPort,\
                                                user=self._dbConfig.username,\
                                                passwd=self._dbConfig.password)
        self._info("Connected")

    def createTable(self, tableName, tableSchemaDict):
        """@brief Create the table in the connected database.
           @param tableName The table we're innterested in.
           @param tableSchemaDict The schema for the table in dict form."""
        if not DatabaseIF.CheckTableExists(self._dbCon, tableName):
            DatabaseIF.CreateTable(self._dbCon, tableName, tableSchemaDict)
            self._info("Created the {} table".format(tableName))
        else:
            raise Exception("The {} table already exists in the database.".format(tableName))

    def ensureTableExists(self, tableName, tableSchemaDict, autoCreate):
        """@brief Check that the table a table exists in the database.
           @param tableName The table we're interested in.
           @param tableSchemaDict The schema for the table in dict form.
           @param autoCreate IF True then auto create the table."""

        if not DatabaseIF.CheckTableExists(self._dbCon, tableName):
            if autoCreate:

                self.createTable(tableName, tableSchemaDict)

            else:

                raise DatabaseIFError("{} database table not found.".format(tableName) )

    def insertRow(self, dataDict, tableName, tableSchemaDict):
        """@brief Insert a row into a database table.
                Must have previously connected to the database.
           @param dataDict The dict holding the table data
           @param tableName The name of the table to insert data into
           @param tableSchemaDict The schema for the database table."""
        insertableDict = DatabaseIF.GetInsertableDict(tableName, dataDict, tableSchemaDict)
        DatabaseIF.InsertRow(self._dbCon, tableName, insertableDict)

    def executeSQL(self, sqlCmd):
        """@brief execute an SQL cmd"""
        self._debug("EXECUTE SQL: {}".format(sqlCmd))
        dictCursor = self._dbCon.cursor(mysqldb.cursors.DictCursor)

        dictCursor.execute(sqlCmd)
        return dictCursor.fetchall()

    def disconnect(self):
        """@brief Disconnect from the database server."""
        try:
            if self._dbCon:
                self._dbCon.close()
                self._dbCon = None
                self._info("Disconnected from {}:{}".format(self._dbConfig.serverAddress, self._dbConfig.serverPort))
        except:
            pass

    def createDatabase(self):
        """@brief Create the database"""
        sql = "CREATE DATABASE `{}`;".format(self._dbConfig.dataBaseName)
        self.executeSQL(sql)
        self._info("Created {} database.".format(self._dbConfig.dataBaseName))

    def dropDatabase(self):
        """@brief Delete the database"""
        sql = "DROP DATABASE `{}`;".format(self._dbConfig.dataBaseName)
        self.executeSQL(sql)
        self._info("Deleted {} database.".format(self._dbConfig.dataBaseName))

    def dropTable(self, tableName):
        """@brief Delete a table in the database
           @param tableName The table to drop."""
        sql = "DROP TABLE `{}`;".format(tableName)
        self.executeSQL(sql)
        self._info("Deleted {} database table.".format(tableName))


#!/usr/bin/env python3

import  socketserver
import  socket
from    struct import pack, unpack
import  json
from    time import  sleep

class JSONNetworking(Exception):
    pass

class JSONServer(socketserver.ThreadingTCPServer):
    """@brief Responsible for accepting tcp connections to receive and send json messages."""

    daemon_threads = True
    allow_reuse_address = True

class JsonServerHandler (socketserver.BaseRequestHandler):

    LEN_FIELD = 4

    @staticmethod
    def DictToJSON(aDict):
        """@brief convert a python dictionary into JSON text
           @param aDict The python dictionary to be converted
           @return The JSON text representing aDict"""
        return json.dumps(aDict)

    @staticmethod
    def JSONToDict(jsonText):
        """@brief Convert from JSON text to a python dict. Throws a ValueError
                  if the text is formatted incorrectly.
           @param jsonText The JSON text to be converted to a dict
           @return a python dict object"""
        return json.loads(jsonText)

    @staticmethod
    def TX(request, theDict):
        """@brief Write the dict to the socket as json text
           @param request The request object passed from the handle() method.
           @theDict The python dictionary to send."""
        if request:
            msg = JsonServerHandler.DictToJSON(theDict).encode()
            sFmt = ">%ds" % len(msg)
            body = pack(sFmt, msg)
            bodyLen = pack('>I', len(body))
            txBuf = bodyLen + body
            request.send(txBuf)

        else:
            raise RuntimeError("TX socket error")

    def handle(self):
        """@brief Handle connections to the server."""
        raise JSONNetworking("!!! You must override this method in a subclass !!!")

class JSONClient(object):

    DEFAULT_RX_POLL_SECS = 0.02
    RX_BUFFER_SIZE       = 2048

    def __init__(self, address, port, keepAliveActSec=1, keepAliveTxSec=15, keepAliveFailTriggerCount=8):
        """@brief Connect to a JSONServer socket.
                  If on a Linux system then the connection will apply a keepalive to the TCP connection.
                  By default this is 2 minutes.
           @param address The address of the JSONServer.
           @param The port on the JSON server to connect to.
           @param keepAliveActSec Activate the keepalive failure this many seconds after it is triggered.
           @param keepAliveTxSec Send a TCP keepalive periodically. This defines the period in seconds.
           @param keepAliveFailTriggerCount Trigger a keepalive failure when this many keepalives fail consecutively."""

        self._socket = socket.socket()
        self._socket.connect( (address, port) )
            
        if hasattr(socket, 'SOL_SOCKET') and\
           hasattr(socket, 'SO_KEEPALIVE') and\
           hasattr(socket, 'IPPROTO_TCP') and\
           hasattr(socket, 'TCP_KEEPIDLE') and\
           hasattr(socket, 'TCP_KEEPINTVL') and\
           hasattr(socket, 'TCP_KEEPCNT'):        
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, keepAliveActSec)
            self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, keepAliveTxSec)
            self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, keepAliveFailTriggerCount)
        self._socket.setblocking(False)
        self._rxBuffer = bytearray()

    def tx(self, theDict, throwError=True):
        """@brief send a python dictionary object to the server via json
           @param theDict The dictionary to send
           @param throwError If True then an exception will be thrown if an error occurs.
                              If False then this method will fail silentley.
           @return True on success. False on failure if throwError = False"""
        try:
            JsonServerHandler.TX(self._socket, theDict)
            return True
        except:
            if throwError:
                raise
            return False

    def _getBodyLen(self, rxBytes):
        """@brief Get the length of the body of the message.
           @param rxBytes The rx buffer containing bytes received.
           @return The length of the body of the message or 0 if we do not have a complete message in the rx buffer."""
        bodyLenFound = 0
        #If we have enough data to extract the length field (start of PDU)
        if len(rxBytes) >= JsonServerHandler.LEN_FIELD:
            # Read the length of the message
            bodyLen = unpack(">I", rxBytes[:JsonServerHandler.LEN_FIELD])[0]
            #If we have the len field + the message body
            if len(rxBytes) >= JsonServerHandler.LEN_FIELD+bodyLen:
                bodyLenFound = bodyLen
        return bodyLenFound

    def _msgAvail(self, rxBytes):
        """@brief Determine is a complete message is present in the rx buffer.
           @param rxBytes The rx buffer containing bytes received.
           @return True if a complete message is present in the RX buffer."""
        msgAvail = False

        bodyLen = self._getBodyLen(rxBytes)

        if bodyLen:
            msgAvail = True

        return msgAvail
        
    def _getDict(self):
        """@brief Get dict from rx data.
           @return The oldest dict in the rx buffer."""
        rxDict = None
        bodyLen = self._getBodyLen(self._rxBuffer)
        #If we have a complete message in the RX buffer
        if bodyLen > 0:
            body = self._rxBuffer[JsonServerHandler.LEN_FIELD:JsonServerHandler.LEN_FIELD+bodyLen]
            rxDict = JsonServerHandler.JSONToDict( body.decode() )
            #Remove the message just received from the RX buffer
            self._rxBuffer = self._rxBuffer[JsonServerHandler.LEN_FIELD+bodyLen:]

        return rxDict
        
    def rx(self, blocking=True, pollPeriodSeconds=DEFAULT_RX_POLL_SECS):
        """@brief Get a python dictionary object from the server.
           @param blocking If True block until complete message is received.
           @param If blocking wait for this period in seconds between checking for RX data.
           @return A received dictionary of None if not blocking and no dictionary is available."""

        while not self._msgAvail(self._rxBuffer):

            try:
                rxd = self._socket.recv(JSONClient.RX_BUFFER_SIZE)
                if len(rxd) > 0:
                    self._rxBuffer = self._rxBuffer + rxd
                else:
                    raise RuntimeError("Socket closed")
    
                if not blocking:
                    break

            except BlockingIOError:
                if not blocking:
                    break

            if not blocking:
                sleep(pollPeriodSeconds)

        return self._getDict()

    def close(self):
        """@brief Close the socket connection to the server."""
        if self._socket:
            self._socket.close()



#!/usr/bin/env python3

import sys
import  queue
from    datetime import datetime
import  itertools
import  threading
from    functools import partial
import  asyncio

from bokeh.server.server import Server
from bokeh.application import Application
from bokeh.application.handlers.function import FunctionHandler
from bokeh.plotting import figure, ColumnDataSource
from bokeh.models import Range1d
from bokeh.palettes import Category20_20 as palette

from bokeh.plotting import save, output_file
from bokeh.layouts import gridplot, column, row
from bokeh.models.widgets import CheckboxGroup
from bokeh.models.widgets.buttons import Button
from bokeh.models.widgets import TextInput
from bokeh.models import TextAreaInput
from bokeh.models import Panel, Tabs
from bokeh.models import DataTable, TableColumn
from bokeh.models import CustomJS
from bokeh import events

class UpdateEvent(object):
    """@brief Responsible for holding the state of an event sent from a non GUI thread
              to the GUI thread context in order to update the GUI. The details of these
              updates will be specific to the GUI implemented. Therefore this class should
              be extended to include the events that are specific to the GUI implemented."""

    UPDATE_STATUS_TEXT = 1 # This is an example of an event. It is intended to be used to
                           # update the status line in the GUI to provide the user with
                           # some feedback as to the current state of the GUI.

    def __init__(self, id, argList=None):
        """@brief Constructor
           @param id An integer event ID
           @param argList A list of arguments associated with the event"""
        #As this is esentially a holding class we don't attempt to indicate provate attributes
        self.id = id
        self.argList = argList

class TimeSeriesPoint(object):
    """@brief Resonsible for holding a time series point on a trace."""
    def __init__(self, traceIndex, value, timeStamp=None):
        """@brief Constructor
           @param traceIndex The index of the trace this reading should be applied to.
                             The trace index starts at 0 for the top left plot (first
                             trace added) and increments with each call to addTrace()
                             on TimeSeriesPlotter instances.
           @param value The Y value
           @param timeStamp The x Value."""
        self.traceIndex = traceIndex
        if timeStamp:
            self.time = timeStamp
        else:
            self.time = datetime.now()
        self.value = value

class TabbedGUI(object):
    """@brief A Generalised class responsible for plotting real time data."""

    @staticmethod
    def GetFigure(title=None, yAxisName=None, yRangeLimits=None, width=400, height=400):
        """@brief A Factory method to obtain a figure instance.
                  A figure is a single plot area that can contain multiple traces.
           @param title The title of the figure.
           @param yAxisName The name of the Y axis.
           @param yRangeLimits If None then the Y azxis will auto range.
                               If a list of two numerical values then this
                               defines the min and max Y axis range values.
           @param width The width of the plot area in pixels.
           @param height The height of the plot area in pixels.
           @return A figure instance."""
        if yRangeLimits and len(yRangeLimits) == 2:
            yrange = Range1d(yRangeLimits[0], yRangeLimits[1])
        else:
            yrange = None

        fig = figure(title=title,
                     x_axis_type="datetime",
                     x_axis_location="below",
                     y_range=yrange,
                     plot_width=width,
                     plot_height=height)
        fig.yaxis.axis_label = yAxisName
        return fig

    def __init__(self, docTitle, bokehPort=9090):
        """@brief Constructor.
           @param docTitle The document title.
           @param bokehPort The port to run the server on."""
        self._docTitle=docTitle
        self._bokehPort=bokehPort
        self._doc = None
        self._tabList = []
        self._server = None

    def stopServer(self):
        """@brief Stop the bokeh server"""
        sys.exit()

    def isServerRunning(self):
        """@brief Check if the server is running.
           @param True if the server is running. It may take some time (~ 20 seconds)
                  after the browser is closed before the server session shuts down."""
        serverSessions = "not started"
        if self._server:
            serverSessions = self._server.get_sessions()

        serverRunning = True
        if not serverSessions:
                serverRunning = False

        return serverRunning

    def runBokehServer(self):
        """@brief Run the bokeh server. This is a blocking method."""
        apps = {'/': Application(FunctionHandler(self.createPlot))}
        self._server = Server(apps, port=self._bokehPort)
        self._server.show("/")
        self._server.run_until_shutdown()

    def _run(self, method, args=[]):
        """@brief Run a method in a separate thread. This is useful when
                  methods are called from gui events that take some time to execute.
                  For such methods the gui callback should call this method to execute
                  the time consuming methods in another thread.
           @param method The method to execute.
           @param A tuple of arguments to pass to the method.
                  If no arguments are required then an empty tuple should be passed."""
        thread = threading.Thread(target=method, args=args)
        thread.start()

    def _sendUpdateEvent(self, updateEvent):
        """@brief Send an event to the GUI context to update the GUI. When methods
                  are executing outside the gui thread but need to update the state
                  of the GUI, events must be sent to the gui context in order to update
                  the gui elements when they have the correct locks.
           @param updateEvent An UpdateEvent instance."""
        self._doc.add_next_tick_callback( partial(self._rxUpdateEvent, updateEvent)  )

    def _rxUpdateEvent(self, updateEvent):
        """@brief Receive an event into the GUI context to update the GUI.
           @param updateEvent An PSUGUIUpdateEvent instance. This method will
                              be specific to the GUI implemented and must therefore
                              be overridden in child classes."""
        raise Exception("BUG: The _rxUpdateEvent() method must be implemented by classes that are children of the TabbedGUI class.")

class TimeSeriesPlotter(TabbedGUI):
    """@brief Responsible for plotting data on tab 0 with no other tabs."""

    def __init__(self, docTitle, bokehPort=9091, topCtrlPanel=True):
        """@Constructor
           @param docTitle The document title.
           @param bokehPort The port to run the server on.
           @param topCtrlPanel If True then a control panel is displayed at the top of the plot.
           """
        super().__init__(docTitle, bokehPort=bokehPort)
        self._statusAreaInput = None
        self._figTable=[[]]
        self._grid = None
        self._topCtrlPanel=topCtrlPanel
        self._srcList = []
        self._colors = itertools.cycle(palette)
        self._queue = queue.Queue()
        self._plottingEnabled = True

    def addTrace(self, fig, legend_label, line_color=None, line_width=1):
        """@brief Add a trace to a figure.
           @param fig The figure to add the trace to.
           @param line_color The line color
           @param legend_label The text of the label.
           @param line_width The trace line width."""
        src = ColumnDataSource({'x': [], 'y': []})

        #Allocate a line color if one is not defined
        if not line_color:
            line_color = next(self._colors)

        fig.line(source=src,
                 line_color = line_color,
                 legend_label = legend_label,
                 line_width = line_width)
        self._srcList.append(src)

    def _update(self):
        """@brief called periodically to update the plot traces."""
        if self._plottingEnabled:
            while not self._queue.empty():
                timeSeriesPoint = self._queue.get()
                new = {'x': [timeSeriesPoint.time],
                       'y': [timeSeriesPoint.value]}
                source = self._srcList[timeSeriesPoint.traceIndex]
                source.stream(new)

    def addValue(self, traceIndex, value, timeStamp=None):
        """@brief Add a value to be plotted. This adds to queue of values
                  to be plotted the next time _update() is called.
           @param traceIndex The index of the trace this reading should be applied to.
           @param value The Y value to be plotted.
           @param timeStamp The timestamp associated with the value. If not supplied
                            then the timestamp will be created at the time when This
                            method is called."""
        timeSeriesPoint = TimeSeriesPoint(traceIndex, value, timeStamp=timeStamp)
        self._queue.put(timeSeriesPoint)

    def addRow(self):
        """@brief Add an empty row to the figures."""
        self._figTable.append([])

    def addToRow(self, fig):
        """@brief Add a figure to the end of the current row of figues.
           @param fig The figure to add."""
        self._figTable[-1].append(fig)

    def createPlot(self, doc, ):
        """@brief create a plot figure.
           @param doc The document to add the plot to."""
        self._doc = doc
        self._doc.title = self._docTitle

        plotPanel = self._getPlotPanel()

        self._tabList.append( Panel(child=plotPanel,  title="Plots") )
        self._doc.add_root( Tabs(tabs=self._tabList) )
        self._doc.add_periodic_callback(self._update, 100)

    def _getPlotPanel(self):
        """@brief Add tab that shows plot data updates."""
        self._grid = gridplot(children = self._figTable, sizing_mode = 'scale_both',  toolbar_location='left')

        if self._topCtrlPanel:
            checkbox1 = CheckboxGroup(labels=["Plot Data"], active=[0, 1],max_width=70)
            checkbox1.on_change('active', self._checkboxHandler)

            self._fileToSave = TextInput(title="File to save", max_width=150)

            saveButton = Button(label="Save", button_type="success", width=50)
            saveButton.on_click(self._savePlot)

            shutDownButton = Button(label="Quit", button_type="success", width=50)
            shutDownButton.on_click(self.stopServer)

            self._statusBarWrapper = StatusBarWrapper()

            plotRowCtrl = row(children=[checkbox1, saveButton, self._fileToSave, shutDownButton])
            plotPanel = column([plotRowCtrl, self._grid, self._statusBarWrapper.getWidget()])
        else:
            plotPanel = column([self._grid])

        return plotPanel

    def _savePlot(self):
        """@brief Save plot to a single html file. This allows the plots to be
                  analysed later."""
        if self._fileToSave and self._fileToSave.value:
            if self._fileToSave.value.endswith(".html"):
                filename = self._fileToSave.value
            else:
                filename = self._fileToSave.value + ".html"
            output_file(filename)
            # Save all the plots in the grid to an html file that allows
            # display in a browser and plot manipulation.
            save( self._grid )
            self._statusBarWrapper.setStatus("Saved {}".format(filename))

    def _checkboxHandler(self, attr, old, new):
        """@brief Called when the checkbox is clicked."""
        if 0 in list(new):  # Is first checkbox selected
            self._plottingEnabled = True
            self._statusBarWrapper.setStatus("Plotting enabled")
        else:
            self._plottingEnabled = False
            self._statusBarWrapper.setStatus("Plotting disabled")

    def runNonBlockingBokehServer(self):
        """@brief Run the bokeh server in a separate thread. This is useful
                  if the we want to load realtime data into the plot from the
                  main thread."""
        self._serverThread = threading.Thread(target=self._runBokehServer)
        self._serverThread.setDaemon(True)
        self._serverThread.start()

    def _runBokehServer(self):
        """@brief Run the bokeh server. This is called when the bokeh server is executed in a thread."""
        apps = {'/': Application(FunctionHandler(self.createPlot))}
        #As this gets run in a thread we need to start an event loop
        evtLoop = asyncio.new_event_loop()
        asyncio.set_event_loop(evtLoop)
        self._server = Server(apps, port=self._bokehPort)
        self._server.start()
        #Show the server in a web browser window
        self._server.io_loop.add_callback(self._server.show, "/")
        self._server.io_loop.start()

class StatusBarWrapper(object):
    """@brief Responsible for presenting a single status line of text in a GUI
              that runs the width of the page (normally at the bottom)."""
    def __init__(self):
        data = dict(
            status = [],
        )
        self.source = ColumnDataSource(data)

        columns = [
                TableColumn(field="status", title="status"),
            ]
        self.statusBar = DataTable(source=self.source, columns=columns, height_policy="fixed", height=50, header_row=False, index_position=None)

    def getWidget(self):
        """@brief return an instance of the status bar widget to be added to a layout."""
        return self.statusBar

    def setStatus(self, msg):
        """@brief Set the message iun the status bar.
           @param The message to be displayed."""
        self.source.data = {"status": [msg]}

class ReadOnlyTableWrapper(object):
    """@brief Responsible for presenting a table of values that can be updated dynamically."""
    def __init__(self, columnNameList, height=400, heightPolicy="auto", showLastRows=0, index_position=None):
        """@brief Constructor
           @param columnNameList A List of strings denoting each column in the 2 dimensional table.
           @param height The hieght of the table viewport in pixels.
           @param heightPolicy The height policy (auto, fixed, fit, min, max). default=fixed.
           @param showLastRows The number of rows to show in the table. If set to 2 then only
                  the last two rows in the table are displayed but they ate scrolled into view.
                  The default=0 which will display all rows and will not scroll the latest
                  into view..
           @param index_position The position of the index column in the table. 0 = the first
                  column. Default is None which does not display the index column."""
        self._columnNameList = columnNameList
        self._dataDict = {}
        self._columns = []
        for columnName in columnNameList:
            self._dataDict[columnName]=[]
            self._columns.append( TableColumn(field=columnName, title=columnName) )

        self._source = ColumnDataSource(self._dataDict)

        self._dataTable = DataTable(source=self._source, columns=self._columns, height=height, height_policy=heightPolicy, frozen_rows=-showLastRows, index_position=index_position)

    def getWidget(self):
        """@brief return an instance of the table widget to be added to a layout."""
        return self._dataTable

    def setRows(self, rowList):
        """@brief Set the rows in the table.
           @param rowList A list of rows of data. Each row must contain a list of values for each column in the table."""
        for _row in rowList:
            if len(_row) != len(self._columnNameList):
                raise Exception("{} row should have {} values.".format(_row, len(self._columnNameList)))
        dataDict = {}
        colIndex = 0
        for columnName in self._columnNameList:
            valueList = []
            for _row in rowList:
                valueList.append( _row[colIndex] )
            dataDict[columnName]=valueList

            colIndex = colIndex + 1
        self._source.data = dataDict

    def appendRow(self, _row):
        """@brief Set the rows in the table.
           @param rowList A list of rows of data. Each row must contain a list of values for each column in the table."""
        dataDict = {}
        colIndex = 0
        for columnName in self._columnNameList:
            valueList = [_row[colIndex]]
            dataDict[columnName]=valueList
            colIndex = colIndex + 1
        self._source.stream(dataDict)

class AlertButtonWrapper(object):
    """@brief Responsible for presenting a button that when clicked displayed an alert dialog."""
    def __init__(self, buttonLabel, alertMessage, buttonType="default", onClickMethod=None):
        """@brief Constructor
           @param buttonLabel The text displayed on the button.
           @param alertMessage The message displayed in the alert dialog when clicked.
           @param buttonType The type of button to display (default, primary, success, warning, danger, light)).
           @param onClickMethod An optional method that is called when the alert OK button has been clicked.
        """
        self._button = Button(label=buttonLabel, button_type=buttonType)
        if onClickMethod:
            self.addOnClickMethod(onClickMethod)

        source = {"msg": alertMessage}
        callback1 = CustomJS(args=dict(source=source), code="""
            var msg = source['msg']
            alert(msg);
        """)
        self._button.js_on_event(events.ButtonClick, callback1)

    def addOnClickMethod(self, onClickMethod):
        """@brief Add a method that is called after the alert dialog has been displayed.
           @param onClickMethod The method that is called."""
        self._button.on_click(onClickMethod)

    def getWidget(self):
        """@brief return an instance of the button widget to be added to a layout."""
        return self._button

class ShutdownButtonWrapper(object):
    """@brief Responsible for presenting a shutdown button. When the button is clicked
              an alert message is displayed instructing the user to close the browser
              window. When the OK button in the alert dialog is clicked the
              application is shutdown."""
    def __init__(self, shutDownMethod):
        """@brief Constructor
           @param shutDownMethod The method that is called to shutdown the application.
        """
        self._alertButtonWrapper = AlertButtonWrapper("Quit",\
                                                      "The application is shutting down. Please close the browser window",\
                                                      buttonType="danger",\
                                                      onClickMethod=shutDownMethod)

    def getWidget(self):
        """@brief return an instance of the shutdown button widget to be added to a layout."""
        return self._alertButtonWrapper.getWidget()



#!/usr/bin/env python3

from queue import Queue

class Conduit(object):
	"""@brief A generalised conduit implementation. A conduit is used for 
	          communication between tow processes. Each Conduit has an A 
	          and a B end. Messages can be pushed into each end and 
	          pulled from each end.	
	"""
	CONDUIT_TYPE_QUEUE 			= 1
	CONDUIT_TYPE_TCP_CONNECTION = 2
	
	def __init__(self, uio=None, cName="", conduitType=CONDUIT_TYPE_QUEUE, readBlock=True, readBlockTimeoutSeconds=None, maxSize = 0, ):
		"""@brief Responsible for providing a conduit for data between entities. 
		   @param uio A User input/output object. If supplied then debug info for the conduit will be recorded.
		   @param cName The conduit name. Only useful if a uio object has been passed for debugging purposes.
		   @param readBlock If true then all getX() methods will block until data is available.
		   @param readBlockTimeoutSeconds The time in seconds for a read (when readBlock=True) to timeout. The default = None (block indefinatley).
		   @param maxSize Maximum number of elements in the queue. Only valid if conduitType = CONDUIT_TYPE_QUEUE."""
		
		if conduitType == Conduit.CONDUIT_TYPE_QUEUE:
			self._conduit = QueueConduit(uio=uio, cName=cName, maxQueueSize=maxSize, readBlock=readBlock, readBlockTimeoutSeconds=readBlockTimeoutSeconds)
			
		elif conduitType == Conduit.CONDUIT_TYPE_TCP_CONNECTION:
			raise Exception("TCP conduits not yet implemented.")
		
		else:
			raise Exception("%d is an invalid conduit type." % (conduitType) )
			
	def putA(self, data):
		"""@brief put some data in the A -> B side conduit.
    	   @param data The data object to be pushed into the conduit."""
		self._conduit.putA(data)
    	
	def putB(self, data):
		"""@brief put some data in the B -> A side conduit.
    	   @param data The data object to be pushed into the conduit."""
		self._conduit.putB(data)
    	
	def getA(self):
		"""@brief Get some data from the B -> A conduit.
		   @return The data from the conduit or None of no data is available."""
		return self._conduit.getA()
						
	def getB(self):
		"""@brief Get some data from the A -> B conduit.
		   @return The data from the conduit or None of no data is available."""
		return self._conduit.getB()
		
	def aReadAvailable(self):
		"""@return True if there is data available to be read from the A side of the conduit."""
		return self._conduit.aReadAvailable()

	def bReadAvailable(self):
		"""@return True if there is data available to be read from the B side of the conduit."""
		return self._conduit.bReadAvailable()

class QueueConduit(Conduit):
	"""@brief Responsible for providing the functionality required to communicate between
	          threads (ITC= Inter Thread Communication).
	          
	          The ITC has an A side and a B side. data can be sent from the A and B sides
	          and is forwarded to the other side."""

	def __init__(self, uio=None, cName="", maxQueueSize = 0, readBlock=True, readBlockTimeoutSeconds=0):
		"""@brief Constructor
		   @param uio A User input/output object. If supplied then debug info for the conduit will be recorded.
		   @param cName The conduit name. Only useful if a uio object has been passed for debugging purposes.
		   @param maxQueueSize The maximum queue size that we will allow (default = 0, no limit)
		   @param readBlock If True then reads will block until data is available or a timeout (if > 0) occurs."""
		  
		self._uio = uio
		self._cName = cName
		self._maxQueueSize 				= maxQueueSize
		self._readBlock     			= readBlock
		self._readBlockTimeoutSeconds 	= readBlockTimeoutSeconds
		self._aToBQueue 				= Queue(maxQueueSize)
		self._bToAQueue	 				= Queue(maxQueueSize)
		
	def _checkQueueSize(self):
		"""@brief check that we have not reached the max queue size."""
		if self._maxQueueSize > 0:
			if self._aToBQueue.qsize() >= self._maxQueueSize:
				raise Exception("%s: A -> B queue full." % (self.__class__.__name__) )
				
			if self._bToAQueue.qsize() >= self._maxQueueSize:
				raise Exception("%s: B -> A queue full." % (self.__class__.__name__) )
	
	def putA(self, data):
		"""@brief put some data in the A -> B side queue.
    	   @param data The data object to be pushed into the queue."""
		self._checkQueueSize()
		
		if self._uio:
			qSize = self._aToBQueue.qsize()
			self._uio.debug("%s: A -> B queue size = %d" % (self._cName, qSize) )
		
		self._aToBQueue.put(data)
    	
	def putB(self, data):
		"""@brief put some data in the B -> A side queue.
    	   @param data The data object to be pushed into the queue."""
		self._checkQueueSize()
		
		if self._uio:
			qSize = self._bToAQueue.qsize()
			self._uio.debug("%s: B -> A queue size = %d" % (self._cName, qSize) )
		
		self._bToAQueue.put(data)
    	
	def getA(self, block=True, timeoutSeconds=0.0):
		"""@brief Get some data from the B -> A queue.
		   @param block If true then the call will block
		   @param timeoutSeconds The time in seconds before exiting (returning None) if no data can be read from the queue.
		   @return The data from the queue or None of no data is available."""
		try:
			return self._bToAQueue.get(block=self._readBlock, timeout=self._readBlockTimeoutSeconds)
		except Queue.Empty:
			return None
						
	def getB(self, block=True, timeoutSeconds=0.0):
		"""@brief Get some data from the A -> B queue.
		   @param block If true then the call will block
		   @param timeoutSeconds The time in seconds before exiting (returning None) if no data can be read from the queue.
		   @return The data from the queue or None of no data is available."""
		try:
			return self._aToBQueue.get(block=self._readBlock, timeout=self._readBlockTimeoutSeconds)
		except Queue.Empty:
			return None
		
	def aReadAvailable(self):
		"""@return True if there is data available to be read from the A side of the queue."""
		return not self._bToAQueue.empty()
	
	def bReadAvailable(self):
		"""@return True if there is data available to be read from the B side of the queue."""
		return not self._aToBQueue.empty()
		   
	
	


			
    	


#!/usr/bin/env python3

import  os
import  socket
from    paramiko import SSHClient, AutoAddPolicy, RSAKey, AuthenticationException, SFTPClient
import  logging
import  threading
import  socketserver
import  select

from    getpass import getuser, getpass

class SSHError(Exception):
    pass

# -------------------------------------------------------------------------------

class ExtendedSSHClient(SSHClient):
    """@brief The ssh client class"""

    CMD_POLL_SECONDS = 0.1

    @staticmethod
    def GetLines(bytes):
        """@brief Split the text into lines.
           @param bytes The bytes containing text to be split into lines of text.
           @return A List of lines of text."""
        text = bytes.decode('utf-8')
        lines = []
        if len(text) > 0:
            elems = text.split("\n")
            lines = ExtendedSSHClient.StripEOL(elems)
        return lines

    @staticmethod
    def StripEOL(lines):
        """@brief Strip the end of line characters from the list of lines of text.
           @param lines A list of lines of text.
           @return The same list of lines of text with EOL removed."""
        noEOLLines = []
        for l in lines:
            l = l.rstrip("\n")
            l = l.rstrip("\r")
            noEOLLines.append(l)
        return noEOLLines

    def __init__(self):
        super(ExtendedSSHClient, self).__init__()
        self.set_missing_host_key_policy(AutoAddPolicy())

    def __exec_command(self, command, bufsize=-1):
        """
        Execute a command on the SSH server.  A new L{Channel} is opened and
        the requested command is executed.  The command's input and output
        streams are returned as python C{file}-like objects representing
        stdin, stdout, and stderr.

        @param command: the command to execute
        @type command: str
        @param bufsize: interpreted the same way as by the built-in C{file()} function in python
        @type bufsize: int
        @return: the stdin, stdout, and stderr of the executing command
        @rtype: tuple(L{ChannelFile}, L{ChannelFile}, L{ChannelFile})

        @raise SSHException: if the server fails to execute the command
        """
        chan = self._transport.open_session()
        chan.exec_command(command)
        stdin = chan.makefile('wb', bufsize)
        stdout = chan.makefile('rb', bufsize)
        stderr = chan.makefile_stderr('rb', bufsize)
        exitStatus = chan.recv_exit_status()
        return stdin, stdout, stderr, exitStatus

    def startCmd(self, command ):
        """
        Start executing a executing a command on the SSH server.  A new L{Channel} is opened and
        the requested command is executed.

        @param command: The command to execute.
        @return The SSH channel session with the command executing in it (it is up to the caller to read data from this
                channel in a timely manner).
        @raise SSHException: if the server fails to execute the command
        """
        chan = self._transport.open_session()
        chan.exec_command(command)
        return chan

    def runCmd(self, cmd, throwError=True):
        """Run a command over an ssh session and return a tuple with the following threee elements
          0 - the return code/exit status of the command
          1 - Lines of text from stdout
          2 - lines of text from stderr

          If throwError is true and the exit status of the copmmand is not 0 then an
          SSHError will be thrown
        """
        stdin, stdout, stderr, exitStatus = self.__exec_command(cmd)
        if throwError and exitStatus != 0:
            errorText = stderr.read()
            if len(errorText) > 0:
                raise SSHError(errorText)
            raise SSHError("The cmd '%s' return the error code: %d" % (cmd, exitStatus))
        return [exitStatus, ExtendedSSHClient.GetLines(stdout.read()), ExtendedSSHClient.GetLines(stderr.read())]

class SSH(object):
    """@brief responsible for connecting an ssh connection, excuting commands."""

    PRIVATE_KEY_FILE_LIST               = ["id_rsa", "id_dsa", 'id_ecdsa']
    PUBLIC_KEY_FILE_LIST                = ["id_rsa.pub", "id_dsa.pub", 'id_ecdsa.pub']
    LOCAL_SSH_CONFIG_PATH               = os.path.join(os.path.expanduser("~"), ".ssh")
    DEFAULT_REMOTE_SSH_AUTH_KEYS_FILE   = "~/.ssh/authorized_keys"
    DROPBEAR_DIR                        = "/etc/dropbear"
    DROPBEAR_AUTH_KEYS_FILE             = "%s/authorized_keys" % (DROPBEAR_DIR)
    SSH_COPY_PROG                       = "/usr/bin/ssh-copy-id"
    SERVER_AUTHORISED_KEYS_FILE         = "~/.ssh/authorized_keys"

    @staticmethod
    def GetPublicKeyFile():
        """@brief Get the public key file from the <HOME>/.ssh"""
        homeFolder = SSH.LOCAL_SSH_CONFIG_PATH
        if not os.path.isdir(homeFolder):
            username = getuser()
            if username == 'root':
                homeFolder = '/root/.ssh'
            else:
                homeFolder = '/home/%s/.ssh' % (username)

        for key in SSH.PUBLIC_KEY_FILE_LIST:
            keyFile = os.path.join(homeFolder, key)
            if os.path.isfile(keyFile):
                return keyFile

        raise SSHError("Unable to find a public key file. Please use the 'ssh-keygen -t rsa' command to generate a key pair.")

    @staticmethod
    def GetPrivateKeyFile():
        """@brief Get the private key file from the <HOME>/.ssh"""
        homeFolder = SSH.LOCAL_SSH_CONFIG_PATH
        if not os.path.isdir(homeFolder):
            username = getuser()
            if username == 'root':
                homeFolder = '/root/.ssh'
            else:
                homeFolder = '/home/%s/.ssh' % (username)

        for key in SSH.PRIVATE_KEY_FILE_LIST:
            keyFile = os.path.join(homeFolder, key)
            if os.path.isfile(keyFile):
                return keyFile

        raise SSHError("Unable to find a public key file. Please use the 'ssh-keygen -t rsa' command to generate a key pair.")

    @staticmethod
    def GetPublicKey():
        """@brief Get the public ssh key from the local machine
           @return The public key."""
        pubKeyFile = SSH.GetPublicKeyFile()

        fd = open(pubKeyFile, 'r')
        lines = fd.readlines()
        fd.close()

        if len(lines) < 1:
            raise SSHError("No public key text found in the %s file on the local computer." % (pubKeyFile))

        publicKey = lines[0]
        publicKey = publicKey.strip('\n')
        publicKey = publicKey.strip('\r')
        publicKey = publicKey.strip()
        return publicKey

    @staticmethod
    def GetSSHKeyAttributes(authKey):
        """@brief Extract the following from an ssh key and return them in a tuple.
           @return a tuple containing

           hostname
           username
           keytype
           key

           If unable to extract the above attributes then None is returned.
        """
        hostname = None

        elems = authKey.split()
        if len(elems) > 2:
            keytype = elems[0]
            key = elems[1]
            tmpElems = elems[2].split("@")
            if len(tmpElems) > 1:
                username = tmpElems[0]
                hostname = tmpElems[1]
            else:
                username = elems[2]
                hostname = "?"

        if hostname != None:
            return (hostname, username, keytype, key)
        return (None, None, None, None)

    def __init__(self, host, username, password=None, useCompression=True, port=22, uio=None, privateKeyFile = None):
        """@brief Constructor
           @param host The SSH hostname
           @param username The ssh username
           @param password The ssh password (default=None)
           @param useCompression If True then use compression on the ssh session (default=True)
           @param port The ssh port number (default = 22)
           @param uio A UIO instance (default=None)
           @param privateKeyFile The private ssh keyfile (default=None=Use default private keyfile)
           """
        self._host              = host
        self._port              = port
        self._username          = username
        self._localAddress      = None
        self.useCompression     = useCompression
        self._password          = password
        self._uio               = uio
        if privateKeyFile:
            self._privateKeyFile    = privateKeyFile
        else:
            self._privateKeyFile = SSH.GetPrivateKeyFile()

        self._sshPrivateKey = RSAKey.from_private_key_file(self._privateKeyFile)

        self._ssh = ExtendedSSHClient()
        logging.getLogger("paramiko").setLevel(logging.WARNING)
        self._ssh.set_missing_host_key_policy(AutoAddPolicy())
        self._sftp = None

    def _info(self, text):
        """@brief Present an info level message to the user.
           @param text The text to be presented to the user."""
        if self._uio:
            self._uio.info(text)

    def _warn(self, text):
        """@brief Present an warning level message to the user.
           @param text The text to be presented to the user."""
        if self._uio:
            self._uio.warn(text)

    def _connect(self, connectSFTPSession=False):
        """@brief Connect the ssh connection
           @param connectSFTPSession If True then just after the ssh connection
                  is built an SFT session will be built ready for file transfer.
           @return a ref to the SSHClient object"""
        if not self._ssh:
            self._ssh = ExtendedSSHClient()

        self._ssh.connect(self._host, username=self._username, password=self._password, port=self._port,
                          pkey=self._sshPrivateKey)
        # It can be usefull to know what local IP address was used to reach the ssh server
        self._localAddress = self._ssh.get_transport().sock.getsockname()[0]
        self._ssh.get_transport().use_compression(self.useCompression)
        if connectSFTPSession:
            self._sftp = SFTPClient.from_transport( self._ssh.get_transport() )
        return self._ssh

    def getLocalAddress(self):
        """@brief Get the local IP address of the network interface used to connect to the ssh server"""
        return self._localAddress

    def getSSHClient(self):
        """@brief return a ref to the SSHClient object"""
        return self._ssh

    def close(self):
        """@brief Close an open ssh connection."""
        if self._ssh:
            self._ssh.close()
            self._ssh = None

        if self._sftp:
            self._sftp = None

    def getTransport(self):
        """@brief Get the ssh transport object. Should only be
                  called when the ssh session is connected.
           @return The ssh transport object."""
        return self._ssh.get_transport()

    def runCmd(self, cmd, throwError = True):
        """@brief Run a command over an ssh session.
           @return A tuple with the following three elements when the command is complete.
          0 - the return code/exit status of the command
          1 - Lines of text from stdout
          2 - lines of text from stderr

          If throwError is true and the exit status of the copmmand is not 0 then an
          SSHError will be thrown
        """
        return self._ssh.runCmd(cmd, throwError=throwError)

    def startCmd(self, cmd):
        """@brief Start executing a command. This will return after starting the command and before the command has completed.
                    The following methods maybe called to interrogate the command executions
                        channel.exit_status_ready()
                        channel.recv_ready()
                            channel.channel.recv(len(channel.in_buffer))
                        channel.recv_stderr_ready()
                            channel.recv_stderr(len(channel.in_stderr_buffer))

                        When the command is complete the caller should call channel.close()
           @return A channel instance in which the command is executing.
           """
        return self._ssh.startCmd(cmd)

    def connect(self, enableAutoLoginSetup=False, connectSFTPSession=False):
        """@brief Connect the ssh connection
           @param enableAutoLoginSetup If True and auto login is not setup the
                  user is prompted for the password and the local ssh public key
                  is copied to the server.
           @param connectSFTPSession If True then just after the ssh connection
                  is built an SFT session will be built ready for file transfer.
           @return True if connected without setting up auto login. False if auto
                   login was setup for connection to succeed."""
        setupAutoLogin = False
        try:
            self._connect(connectSFTPSession=connectSFTPSession)

        except AuthenticationException:
            self._setupAutologin()
            self._connect(connectSFTPSession=connectSFTPSession)
            setupAutoLogin = True
        return setupAutoLogin

    def _setupAutologin(self):
        """Setup autologin on the ssh server."""

        publicKeyFilename = "{}.pub".format(self._privateKeyFile)
        publicKeyFile = os.path.join(SSH.LOCAL_SSH_CONFIG_PATH, publicKeyFilename)
        if not os.path.isfile(publicKeyFile):
            raise SSHError(
                "{} public key file not found. Please create a public/private key pair and try again.".format(
                    publicKeyFile))

        self._warn("Auto login to the ssh server failed authentication.")
        self._info("Copying the local public ssh key to the ssh server for automatic login.")
        self._info("Please enter the ssh server ({}) password for the user: {}".format(self._host, self._username))

        self._password = self._uio.getPassword("SSH password: ")

        self._connect()

        self._ensureAutoLogin()

        self.close()

        self._info("Local public ssh key copied to the ssh server.")

    def _ensureAutoLogin(self):
        """@brief Ensure that ssh auto login is enabled."""

        localPublicKey = SSH.GetPublicKey()
        _hostname, _username, _keytype, _ = SSH.GetSSHKeyAttributes(localPublicKey)
        if _hostname == None:
            _hostname = socket.gethostname()
        if _username == None:
            _username = getpass.getuser()

        self._info("Using key: %s@%s" % (_username, _hostname))
        remoteAuthorisedKeys = self.getRemoteAuthorisedKeys()
        # Check to see if the remote authorised keys contains the local public key
        updateAuthKeys = True
        for remoteAuthorisedKeys in remoteAuthorisedKeys:
            if remoteAuthorisedKeys.find(localPublicKey) == 0:
                updateAuthKeys = False
                break

        if updateAuthKeys:
            remoteAuthKeysFile = self.updateAuthorisedKeys(localPublicKey)
            self._info("Updated the remote %s file from the local %s file." % (remoteAuthKeysFile, SSH.GetPublicKeyFile()))
        else:
            self._info("The server already has the local ssh key (%s) in its authorized_key file." % (SSH.GetPublicKeyFile()))

    def updateAuthorisedKeys(self, publicKey):
        """Update the authorised keys file on the remote ssh server with the
           public ssh key"""
        authKeysFile = self.getRemoteAuthorisedKeyFile()
        cmd = "test -d %s" % (authKeysFile)
        rc, stdoutlines, stderrlines = self.runCmd(cmd, throwError=False)
        if rc == 0:
            # If this is a dir with nothing in it, delete it and create an empty
            # authorized_keys file.
            self.runCmd("rmdir %s" % (authKeysFile))
            self.runCmd("touch %s" % (authKeysFile))

        self.runCmd("echo \"%s\" >> %s" % (publicKey, authKeysFile))
        self.runCmd("chmod 600 %s" % (authKeysFile))
        return authKeysFile

    def getRemoteAuthorisedKeys(self):
        """Get the remote authorised keys file over the ssh connection."""
        authKeysFile = self.getRemoteAuthorisedKeyFile()
        cmd = "test -f %s" % (authKeysFile)
        rc, stdoutLines, stderrLines = self.runCmd(cmd, throwError=False)
        if rc != 0:
            # Auth keys file not found, attempt to create an empty one.
            cmd = "touch %s" % (authKeysFile)
            rc, stdoutLines, stderrLines = self.runCmd(cmd, throwError=False)
            cmd = "test -f %s" % (authKeysFile)
            rc, stdoutLines, stderrLines = self.runCmd(cmd, throwError=False)
            if rc != 0:
                raise SSHError("!!! Server auth keys file not found (%s). Failed to create it." % (authKeysFile))

        rc, stdoutLines, stderrLines = self.runCmd("cat %s" % (authKeysFile), throwError=False)

        # Ensure we only return non empty lines
        authKeyLines = []
        for l in stdoutLines:
            if len(l.strip()) > 0:
                authKeyLines.append(l)
        return authKeyLines

    def getRemoteAuthorisedKeyFile(self):
        """@brief Return the remote authorised key file for the current ssh connection."""

        authKeysFile = SSH.DEFAULT_REMOTE_SSH_AUTH_KEYS_FILE

        cmd = "test -d %s" % (SSH.DROPBEAR_DIR)
        rc, stdoutLines, stderrLines = self.runCmd(cmd, throwError=False)
        if rc == 0:
            authKeysFile = SSH.DROPBEAR_AUTH_KEYS_FILE

        return authKeysFile

    def getFile(self, remoteFilePath, localFilePath ):
        """@brief Get a file from the sftp server
           @param remoteFilePath The remote file on the ssh server.
           @param localFilePath The path of the file after it's been received"""
        if self._sftp:
            self._sftp.get(remoteFilePath,localFilePath)
        else:
            raise SSHError("SFTP not connected.")

    def putFile(self, localFilePath, remoteFilePath ):
        """@brief Get a file from the sftp server
           @param localFilePath The path of the file after it's been received
           @param remoteFilePath The remote file on the ssh server."""
        if self._sftp:
            self._sftp.put(localFilePath, remoteFilePath)
        else:
            raise SSHError("SFTP not connected.")

    def getAuthKeyBackupFile(self, maxBackupFileCount=10):
        """@brief Get the name of the backup name for the authorised keys file.
           @param maxBackupFileCount The maximum number of backup files to keep.
           @return None
           - We create up to maxBackupFileCount backup files.
           - Once all the backup files have been created we always replace the oldest
             backup file.
           - The backup files have the suffix .backup1, .backup2 etc.
        """
        authKeysFile = self.getRemoteAuthorisedKeyFile()
        authKeyBackupFilePart = "%s.backup" % (authKeysFile)
        suffixNum = 1
        while True:
            authKeyBackupFileName = "%s%d" % (authKeyBackupFilePart, suffixNum)
            rc, stdoutLines, stderrLines = self.runCmd("test -f %s" % (authKeyBackupFileName), throwError=False)
            if rc != 0:
                return authKeyBackupFileName
            # If all the backup files have been created
            if suffixNum >= maxBackupFileCount:
                # List the files in creation order (oldest first)
                rc, stdoutLines, stderrLines = self.runCmd("ls -ltr %s*" % (authKeyBackupFilePart), throwError=False)
                if rc == 0:
                    if len(stdoutLines) > 0:
                        elems = stdoutLines[0].split()
                        if len(elems) > 0:
                            # Return the oldest file as the next backup filename so that we roll
                            # around the always replacing the oldest backup file.
                            backupfile = elems[len(elems) - 1]
                            return backupfile

                raise SSHError("Unable to %s to %s. Please manually remove the backup files on the ssh server." % (
                authKeysFile, authKeyBackupFilePart))

            suffixNum = suffixNum + 1

    def _getPublicKeyID(self, publicKey):
        """@brief Get the public key ID string.
           @param publicKey The ssh public key string"""
        elems = publicKey.split()
        if len(elems) == 3:
            return elems[2]
        raise SSHError("{} is an invalid public key.".format(publicKey) )

    def removeAuthKey(self, publicKey):
        """@brief Remove authorised keys from the server authorised keys file.
           @param publicKeysForRemoval A list of public keys for removal."""
        remove = False
        previousAuthKeysFile = self.getAuthKeyBackupFile()
        publicKeyID = self._getPublicKeyID(publicKey)
        self._info("Public ssh key ID: {}".format(publicKeyID))
        authKeysFile = self.getRemoteAuthorisedKeyFile()
        tmpAuthKeysFile = "%s.tmp" % (authKeysFile)
        retCode, publicKeyList, _ = self.runCmd("cat {}".format(authKeysFile), throwError=False)
        if retCode == 0:
            newAuthKeysList = []
            for publicKey in publicKeyList:
                if publicKey.find(publicKeyID) >= 0:
                    self._info("Found {} public key on ssh server.".format(publicKeyID))
                    remove = True
                else:
                    newAuthKeysList.append(publicKey)

        if remove:
            # Remove any pre existing tmp auth keys file
            self.runCmd("rm -f %s" % (tmpAuthKeysFile), throwError=False)

            # Create empty tmp auth keys file
            self.runCmd("touch %s" % (tmpAuthKeysFile))

            for newAuthKey in newAuthKeysList:
                self.runCmd("echo \"%s\" >> %s" % (newAuthKey, tmpAuthKeysFile), throwError=False)

            # Remove any pre existing previous auth keys file
            self.runCmd("rm -f %s" % (previousAuthKeysFile), throwError=False)

            # Move the current auth keys file to the old one and the tmp to the current one
            self.runCmd("mv %s %s" % (authKeysFile, previousAuthKeysFile))
            self.runCmd("mv %s %s" % (tmpAuthKeysFile, authKeysFile))
            self._info("Removed {} public key from the ssh server.".format(publicKeyID))
        else:
            self._info("{} public key not found in ssh server authorised keys file.".format(publicKeyID))

        return remove

class SSHTunnelManager(object):
    """@brief Responsible for setting up, tearing down and maintaining lists of
              SSH port forwarding and ssh reverse port forwarding connections."""

    RX_BUFFER_SIZE = 4096

    def __init__(self, uio, ssh, useCompression):
        """@brief Constructor
           @param uio  UIO instance
           @param ssh An instance of SSHClient that has previously been
                      connected to an ssh server"""
        self._uio = uio
        self._ssh = ssh
        self._useCompression = useCompression
        if not self._ssh.getTransport().is_active():
            raise SSHError("!!! The ssh connection is not connected !!!")

        self._forwardingServerList = []
        self._reverseSShDict = {}

    def startFwdSSHTunnel(self, serverPort, destHost, destPort):
        """@brief Start an ssh port forwarding tunnel. This is a non blocking method.
                  A separate thread will be started to handle data transfer over the
                  ssh forwarding connection.
           @param serverPort The TCP server port. On a port forwarding connection
                             the TCP server runs on the src end of the ssh connection.
                             This is the machine that this python code is executing on.
           @param destHost   The host address of the tunnel destination at the remote
                             end of the ssh connection.
           @param destPort   The host TCP port of the tunnel destination at the remote
                             end of the ssh connection."""
        self._uio.info("Forwarding local TCP server port (%d) to %s:%d on the remote end of the ssh connection." % (
        serverPort, destHost, destPort))
        transport = self._ssh.getTransport()

        class SubHander(ForwardingHandler):
            chain_host = destHost
            chain_port = destPort
            ssh_transport = transport
            ssh_transport.use_compression(self._useCompression)
            uo = self._uio

        forwardingServer = ForwardingServer(('', serverPort), SubHander)
        self._forwardingServerList.append(forwardingServer)
        newThread = threading.Thread(target=forwardingServer.serve_forever)
        newThread.setDaemon(True)
        newThread.start()

    def stopFwdSSHTunnel(self, serverPort):
        """@brief stop a previously started ssh port forwarding server
           @param serverPort The TCP server port which is currently accepting
                             port forwarding connections on."""
        for forwardingServer in self._forwardingServerList:
            forwardingServerPort = forwardingServer.server_address[1]
            if forwardingServerPort == serverPort:
                forwardingServer.shutdown()
                forwardingServer.server_close()
                self._uio.info("Shutdown ssh port forwarding connection using local server port %d." % (serverPort))

    def stopAllFwdSSHTunnels(self):
        """@brief Stop all previously started ssh port forwarding servers.."""
        for forwardingServer in self._forwardingServerList:
            forwardingServer.shutdown()
            forwardingServer.server_close()
            self._uio.info("Shutdown ssh port forwarding on %s." % (str(forwardingServer.server_address)))

    def startRevSSHTunnel(self, serverPort, destHost, destPort):
        """@brief Start an ssh reverse port forwarding tunnel
           @param serverPort The TCP server port. On a reverse port forwarding connection
                             the TCP server runs on the dest end of the ssh connection.
                             This is the machine at the remote end of the ssh connection.
           @param destHost   The host address of the tunnel destination at the local
                             end of the ssh connection.
           @param destPort   The host TCP port of the tunnel destination at the local
                             end of the ssh connection."""
        self._uio.info("Forwarding (reverse) Remote TCP server port (%d) to %s:%d on this end of the ssh connection." % (
        serverPort, destHost, destPort))
        # We add the None refs as the placeholders will be used later
        chan = None
        sock = None
        self._reverseSShDict[serverPort] = (destHost, destPort, chan, sock)

        self._ssh.getTransport().use_compression(self._useCompression)
        self._ssh.getTransport().request_port_forward('', serverPort, handler=self._startReverseForwardingHandler)

    def stopRevSSHTunnel(self, serverPort):
        """@brief stop a previously started reverse ssh port forwarding server
           @param serverPort The TCP server port which is currently accepting
                             port forwarding connections on."""
        if serverPort in self._reverseSShDict:
            revSSHParams = self._reverseSShDict[serverPort]
            chan = revSSHParams[2]
            sock = revSSHParams[3]

            if chan:
                chan.close()

            if sock:
                sock.close()

            self._uio.info("Shutdown reverse ssh port forwarding connection using remote server port %d." % (serverPort))

    def stopAllRevSSHTunnels(self):
        """@brief Stop all previously started reverse ssh port forwarding servers."""
        for key in list(self._reverseSShDict.keys()):
            revSSHParams = self._reverseSShDict[key]
            chan = revSSHParams[2]
            sock = revSSHParams[3]

            if chan:
                chan.close()

            if sock:
                sock.close()

            self._uio.info("Shutdown reverse ssh port forwarding connection using remote server port %d." % (key))

    def stopAllSSHTunnels(self):
        """@brief Stop all ssh tunnels."""
        self.stopAllFwdSSHTunnels()
        self.stopAllRevSSHTunnels()

    # !!! The following methods are internal and should noit be called externally.
    def _getDestination(self, serverPort):
        """@brief Get destination (address and port) for the given server port.
           @param serverPort The TCP server port on the ssh server."""
        if serverPort in self._reverseSShDict:
            revSSHParams = self._reverseSShDict[serverPort]
            return (revSSHParams[0], revSSHParams[1])

        return None

    def _startReverseForwardingHandler(self, chan, xxx_todo_changeme, xxx_todo_changeme1):
        """@brief Called when a channel is connected in order to start a handler thread fot it."""
        (origin_addr, origin_port) = xxx_todo_changeme
        (server_addr, serverPort) = xxx_todo_changeme1
        destHost, destPort = self._getDestination(serverPort)

        hThread = threading.Thread(target=self._reverseForwardingHandler, args=(chan, serverPort, destHost, destPort))
        hThread.setDaemon(True)
        hThread.start()

    def _reverseForwardingHandler(self, chan, serverPort, destHost, destPort):
        """@brief Handle a reverse ssh forwarding connection.
           @param chan A connected channel over an ssh connection.
           @param serverPort The server port (on remote ssh server) from where the reverse ssh connection originated.
           @param destHost The destination host address.
           @param destPort The destination port address."""

        sock = socket.socket()
        # Add references to the chnl and sock so that they can be closed if required
        self._reverseSShDict[serverPort] = (destHost, destPort, chan, sock)
        try:
            sock.connect((destHost, destPort))
        except Exception as e:
            self._uio.error('Forwarding (reverse) request to %s:%d failed: %r' % (destHost, destPort, e))
            return

        self._uio.info('Connected!  Reverse tunnel open %r -> %r -> %r' % (chan.origin_addr,
                                                                          chan.getpeername(), (destHost, destPort)))
        while True:
            r, w, x = select.select([sock, chan], [], [])
            if sock in r:
                data = sock.recv(SSHTunnelManager.RX_BUFFER_SIZE)
                if len(data) == 0:
                    break
                try:
                    chan.send(data)
                except:
                    break
            if chan in r:
                data = chan.recv(SSHTunnelManager.RX_BUFFER_SIZE)
                if len(data) == 0:
                    break
                try:
                    sock.send(data)
                except:
                    break
        chan.close()
        sock.close()
        self._uio.info('Tunnel closed from server port %d' % (serverPort))


class ForwardingServer(socketserver.ThreadingTCPServer):
    """@brief Server responsible for ssh port forwarding"""
    daemon_threads = True
    allow_reuse_address = True


class ForwardingHandler(socketserver.BaseRequestHandler):
    """@brief handler for ssh port forwarding connections."""

    def handle(self):
        try:
            chan = self.ssh_transport.open_channel('direct-tcpip',
                                                   (self.chain_host, self.chain_port),
                                                   self.request.getpeername())
        except Exception as e:
            self.uo.error('Incoming request to %s:%d failed: %s' % (self.chain_host,
                                                                    self.chain_port,
                                                                    repr(e)))
            return
        if chan is None:
            self.uo.error('Incoming request to %s:%d was rejected by the SSH server.' %
                          (self.chain_host, self.chain_port))
            return

        self.uo.info('Connected!  Tunnel open %r -> %r -> %r' % (self.request.getpeername(),
                                                                 chan.getpeername(),
                                                                 (self.chain_host, self.chain_port)))
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(SSHTunnelManager.RX_BUFFER_SIZE)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(SSHTunnelManager.RX_BUFFER_SIZE)
                if len(data) == 0:
                    break
                self.request.send(data)

        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        self.uo.info('Tunnel closed from %r' % (peername,))



#!/usr/bin/env python3.8

import  platform
from    subprocess import call, check_output
import  socket
import  struct

class NetIF(object):
    """@Responsible for determining details about the network interfaces on a PC."""
    
    LINUX_OS_NAME      = "Linux"
    SUPPORTED_OS_NAMES = [ LINUX_OS_NAME ]
    
    ID_STR1            = "inet addr:"
    ID_STR2            = "addr:"
    ID_STR3            = "mask:"
    ID_STR4            = "inet "

    IP_NETMASK_SEP     = "/"
    
    @staticmethod
    def IsAddressInNetwork(ip, net):
        """@brief Determine if an IP address is in a IP network.
           @param ip The IP address of the interface in the format 192.168.1.20
           @param net The network in the format 192.168.1.0/24"""
        ipaddr = socket.inet_aton(ip)
        netaddr, netmask = net.split(NetIF.IP_NETMASK_SEP)
        netaddr = socket.inet_aton(netaddr)

        ipint = struct.unpack("!I", ipaddr)[0]
        netint = struct.unpack("!I", netaddr)[0]
        maskint = (0xFFFFFFFF << (32 - int(netmask))) & 0xFFFFFFFF

        return ipint & maskint == netint

                
    @staticmethod            
    def IPStr2int(addr):  
        """@brief Convert an IP address string to an integer.
           @param addr The IP address string.
           @return The integer represented by the IP address."""                                                             
        return struct.unpack("!I", socket.inet_aton(addr))[0]                       
    
    @staticmethod      
    def Int2IPStr(addr):                                                               
        """@brief Convert an integer to an IP address string.
           @param addr The Integer value of the IP address.
           @return The string value (dotted quad) of the IP address."""
        return socket.inet_ntoa(struct.pack("!I", addr))  

    @staticmethod
    def NetmaskToBitCount(netmask):
        """@brief Convert a dotted quad netmask to a bit count netmask format.
           @param netmask: The dotted quad netmask (E.G 255.255.255.0)
           @return: The netmask as a count of the set bits (E.G 24).
        """
        return sum([bin(int(x)).count('1') for x in netmask.split('.')])

    @staticmethod
    def BitCountToNetMask(bitCount):
        """@brief Convert a bit count to a netmask string
           @param bitCount The number of bits in the netmask.
           @return The netmask string E.G 255.255.255.0"""
        if bitCount > 32:
            raise Exception("{} greater than 32".format(bitCount))
        num=1
        for _ in range(0,bitCount):
            num=num>>1
            num=num|0x80000000
        return NetIF.Int2IPStr(num)
        
    def __init__(self):
        
        self._osName = platform.system()

        self._checkSupportedOS()
        
        self._ifDict = None
        
    def getIFDict(self, readNow=False, includeNoIPIF = False):
        """@param readNow If True the read the current network interface 
                          state now regardless of wether we have read it previously.
                          If False and we have read the state of the network 
                          interfaces previously then return the previous state.
           @param includeNoIPIF If True then interfaces with no IPaddress are included in the results.
           @return A dict of the IP network interfaces on this platform.
                       key   = name of interface
                       value = A list of <IP ADDRESS>/<NET MASK BIT COUNT>"""
        if self._ifDict and not readNow:
            return self._ifDict
                   
        if self._osName.find(NetIF.LINUX_OS_NAME) >= 0 :
            self._ifDict = self.getLinuxIFDict(includeNoIPIF=includeNoIPIF)
                   
        return self._ifDict
    
    def getLinuxIFDict(self, includeNoIPIF = False):
        """@brief Get a dic that contains the local interface details.
           @param includeNoIPIF If True then interfaces with no IPaddress are included in the results.
           @return A dict of the IP network interfaces on this platform.
                       key   = name of interface
                       value = A list of <IP ADDRESS>/<NET MASK BIT COUNT>"""
        netIFDict = {}
        ifName = None
        
        cmdOutput = check_output(['/sbin/ip','a'] ).decode()
        netIFDict = self._getLinuxIFDict(cmdOutput, includeNoIPIF=includeNoIPIF)
        return netIFDict

    def _getLinuxIFDict(self, cmdOutput, includeNoIPIF=False):
        """@brief Get a dic that contains the local interface details.
           @param cmdOutput The ip a command output.
           @param includeNoIPIF If True then interfaces with no IPaddress are included in the results.
           @return A dict of the IP network interfaces on this platform.
                       key   = name of interface
                       value = A list of <IP ADDRESS>/<NET MASK BIT COUNT>"""
        netIFDict = {}
        lines = cmdOutput.lower().split('\n')
        interfaceID = 1
        ifName = None
        ifAddressList = []
        for line in lines:
            elems = line.split(":")
            try:
                #If first element is the IF ID
                #Note that the IF ID may not be sequential
                int(elems[0])
                #Extract the if name
                if len(elems) > 1:
                    ifName = elems[1].strip()
                    ifName = ifName.replace(":", "")
                    ifAddressList = []
                    if includeNoIPIF:
                        netIFDict[ifName] = ifAddressList
            except:
                line = line.strip()
                if line.startswith("inet "):
                    elems = line.split()
                    if len(elems) > 1:
                        ipAddress = elems[1]
                        ifAddressList.append(ipAddress)
                        netIFDict[ifName] = ifAddressList

        return netIFDict

    def getIFName(self, ipAddress):
        """@brief Get the name of the interface which the ip address can be reached directly (not via an IP gateway)
           @param ipAddress The IP address that should be out on an interface.
           @return The dict entry for the interface.
                   key == Interface name
                   value = Interface address.
           @return The name of the local network interface or None if not found."""
        ifDict = self.getIFDict()
        ifNames = list(ifDict.keys())
        ifNames.sort()
        for ifName in ifNames:
            ipDetailsList = ifDict[ifName]
            for ipDetails in ipDetailsList:
                elems = ipDetails.split(NetIF.IP_NETMASK_SEP)
                if len(elems) == 2:
                    ipAddr     = elems[0]
                    netMaskBits= int(elems[1])
                    netMask = NetIF.BitCountToNetMask(netMaskBits)
                    ipAddrInt  = NetIF.IPStr2int(ipAddr)
                    netMaskInt = NetIF.IPStr2int(netMask)
                    network = ipAddrInt&netMaskInt
                    networkStr = socket.inet_ntoa(struct.pack("!I", network))
                    netMaskInt=24
                    if NetIF.IsAddressInNetwork(ipAddress, "%s/%s" % ( networkStr, netMaskInt) ):
                        return ifName

        return None
    
    def _getIFDetails(self, ifName):
        """@brief Get the details of the network interface on this machine given the interface name.
           @param ifName The name of the interface.
           @return A tuple containing the tuples that contain
                  0: IP address
                  1: Netmask"""
        self.getIFDict()
        if ifName in self._ifDict:
            ifDetailsList = self._ifDict[ifName]
            for ifDetails in ifDetailsList:
                if ifDetails:
                    elems = ifDetails.split(NetIF.IP_NETMASK_SEP)
                    if len(elems) == 2:
                        return elems
        return None
    
    def getIFIPAddress(self, ifName):
        """@brief Get the IP address of the network interface on this machine given the interface name.
           @param ifName The name of the interface.
           @return The IP address oft heinterface of None if unknown."""
        ifDetails = self._getIFDetails(ifName)
        if ifDetails and len(ifDetails) == 2:
            return ifDetails[0]
            
        return None
    
    def getIFNetmask(self, ifName):
        """@brief Get the netmask of the network interface on this machine given the interface name.
           @param ifName The name of the interface.
           @return The IP address of the interface of None if unknown."""
        ifDetails = self._getIFDetails(ifName)
        if ifDetails and len(ifDetails) == 2:
            netMaskBitCount = int(ifDetails[1])
            return NetIF.BitCountToNetMask(netMaskBitCount)
            
        return None
    
    def _checkSupportedOS(self):
        """@brief Check that the OS is supported."""
        for supportedOSName in NetIF.SUPPORTED_OS_NAMES:
            if self._osName.find(supportedOSName) != -1:
                return
            
        raise Exception("%s: %s is an unsupported platform", type(self).__name__, self._osName)


#!/usr/bin/python3

import  paho.mqtt.client as mqtt
from    abc import abstractmethod
import  json
import  traceback
from    threading import Condition
from    queue import Queue

class MQTTError(Exception):
  pass

class MQTTRPCClient(object):

    DEFAULT_HOST                = "localhost"
    DEFAULT_PORT                = 1883
    DEFAULT_KEEPALIVE_SECONDS   = 60
    DEFAULT_SRV_ID              = 1

    CLIENT_ID_DICT_KEY          = "SRC"
    METHOD_DICT_KEY             = "METHOD"
    ARGS_DICT_KEY               = "ARGS"
    RESPONSE_DICT_KEY           = "RESPONSE"

    @staticmethod
    def GetServerRPCTopic(idNumber=1):
        """@brief Responsible for getting the Server RPC topic string.
           @param idNumber The idNumber of the RPC server that RPC messages can be sent to.
           @return A string that represents the RPC server topic."""
        return "server%d/RPC" % (idNumber)

    @staticmethod
    def GetClientID(id=1):
        """@brief Responsible for getting a client ID string.
           @param id The unique id of the MQTT client.
           @return A string that represents the unique client ID."""
        return "client%s/RPC" % ( str(id) )

    def __init__(self, uo, options):
        """@brief Responsible fopr providing RPC functionality via an MQTT server.
           @param uo A UO instance that implements the info(text) and error(text) methods
                       to send information to the user.
           @param options An options instance.
                    options.sid = A number to uniquely identify the server
                    options.cid = A number to uniquely identify the client
                    options.server = The host address of the MQTT server
                    options.port = The port number to connect to the MQTT server
                    options.keepalive = The keepalive period in seconds

           The MQTTRPCClient class can only be a base class and should be extended. Child classes must
           implement/override the _onMessage() method."""
        self._uo = uo
        self._options = options

        self._serverRPCTopic = MQTTRPCClient.GetServerRPCTopic(idNumber=self._options.sid)

        self._client = mqtt.Client()
        self._client.on_connect = self._onConnect
        self._client.on_message = self._onMessage

    def connect(self):
        """@brief Connect to the MQTT server"""
        self._client.connect(self._options.server, self._options.port, self._options.keepalive)

    def loopForever(self):
        """@brief Wait here to service messages after a successfll connection."""
        self._client.loop_forever()

    def _onConnect(self, client, userdata, flags, rc):
        """@brief Called when connected to the MQTT server."""
        self._client.subscribe( self._serverRPCTopic )
        self._uo.info("Connected to MQTT server (%s:%d) and subscribed to %s" % (self._options.server, self._options.port, self._serverRPCTopic) )

    @abstractmethod
    def _onMessage(self, client, userdata, msg):
        """@brief Called when a message is received from the MQTT server.
                  This method mst be implented in a child class."""
        pass

class MQTTRPCProviderClient(MQTTRPCClient):

    def __init__(self, uo, options, rpcMethodProviderList):
        """@brief Responsible for providing RPC functionality via an MQTT server.
                  MQTTRPCProviderClient instances implement the RPC at the destination/remote side
                  and return the results to the caller side.
           @param uo A UO instance that implements the info(text) and error(text) methods
                     to send information to the user.
           @param options An options instance.
                    options.sid = A number to uniquely identify the server
                    options.cid = A number to uniquely identify the client receiving responses from the server.
                    options.server = The host address of the MQTT server
                    options.port = The port number to connect to the MQTT server
                    options.keepalive = The keepalive period in seconds
           @param rpcMethodProviderList A list/tuple of class instances that provide the RPC methods."""
        MQTTRPCClient.__init__(self, uo, options)
        self._rpcMethodProviderList=rpcMethodProviderList

    def _onMessage(self, client, userdata, msg):
        """@brief Called when a message is received from the MQTT server.
           @param client:     the client instance for this callback
           @param userdata:   the private user data as set in Client() or userdata_set()
           @param message:    an instance of MQTTMessage.
                              This is a class with members topic, payload, qos, retain."""
        try:

          rxStr = msg.payload.decode()
          self._uo.info( "RX: %s" % (rxStr) )

          jsonDict = json.loads( msg.payload.decode() )

          if MQTTRPCClient.CLIENT_ID_DICT_KEY in jsonDict and MQTTRPCClient.METHOD_DICT_KEY in jsonDict:
              clientID = jsonDict[MQTTRPCClient.CLIENT_ID_DICT_KEY]
              methodName = jsonDict[MQTTRPCClient.METHOD_DICT_KEY]
              args = jsonDict[MQTTRPCClient.ARGS_DICT_KEY]

              method = None
              for rpcMethodProvider in self._rpcMethodProviderList:
                  if hasattr(rpcMethodProvider, methodName):
                      method = getattr(rpcMethodProvider, methodName)

              if method:
                  if len(args) > 0:
                      response = method(args)
                  else:
                      response = method()

                  jsonDict[MQTTRPCClient.RESPONSE_DICT_KEY]=response

                  jsonString = json.dumps(jsonDict)

                  self._uo.info( "TX: %s" % (jsonString) )
                  client.publish(clientID, jsonString)

              else:
                  raise AttributeError("Unable to find a provider of the %s method." % (methodName) )

        except:
            tb = traceback.format_exc()
            self._uo.error(tb)



class MQTTRPCCallerClient(MQTTRPCClient):

    def __init__(self, uo, options, responseTimeoutSeconds = 10):
        """@brief Responsible fopr providing RPC functionality via an MQTT server.
                  MQTTRPCCallerClient instances implement the RPC at the src/caller side
                  sending RPC requests to destination/remote and reciving responses.
           @param uo A UO instance that implements the info(text) and error(text) methods
                     to send information to the user.
           @param options An options instance.
                    options.sid = A number to uniquely identify the server responding to RPC calls.
                    options.cid = A number to uniquely identify the client receiving responses from the server.
                    options.server = The host address of the MQTT server
                    options.port = The port number to connect to the MQTT server
                    options.keepalive = The keepalive period in seconds
           @param timeoutSeconds The number of seconds to wait for a timeout when no response is received."""
        MQTTRPCClient.__init__(self, uo, options)
        self._responseTimeoutSeconds = responseTimeoutSeconds

        self._clientSRC = MQTTRPCClient.GetClientID(options.cid)

        self._serverRPCTopic = MQTTRPCClient.GetServerRPCTopic(idNumber=self._options.sid)

        self._sendMsgDict = None
        self._responseQueue = Queue()

        self._condition = Condition()

    def _onMessage(self, client, userdata, msg):
        """@brief Called when a message is received from the MQTT server.
           @param client:     the client instance for this callback
           @param userdata:   the private user data as set in Client() or userdata_set()
           @param message:    an instance of MQTTMessage.
                              This is a class with members topic, payload, qos, retain.
           @return None"""
        try:

          responseDict = json.loads( msg.payload.decode() )
          if responseDict and MQTTRPCClient.RESPONSE_DICT_KEY in responseDict:
              self._condition.acquire()
              self._responseQueue.put(responseDict[MQTTRPCClient.RESPONSE_DICT_KEY])
              self._condition.notify()
              self._condition.release()

        except:
            tb = traceback.format_exc()
            self._uo.error(tb)

    def _getResponse(self):
        """@brief Internal method used to get the response to an RPC
           @return The response to the RPC or None if a timeout occurs."""

        # blocks until the response is processed or a timeout occurs
        self._condition.wait(timeout=self._responseTimeoutSeconds)

        response = None
        if not self._responseQueue.empty():
            response = self._responseQueue.get()

        return response

    def _getRPCString(self):
        """@brief Get the last RPC called as a string.
           @return The RPC string representation."""
        if self._sendMsgDict and MQTTRPCClient.METHOD_DICT_KEY in self._sendMsgDict and MQTTRPCClient.ARGS_DICT_KEY in self._sendMsgDict:
            return "%s(%s)" % (self._sendMsgDict[MQTTRPCClient.METHOD_DICT_KEY], self._sendMsgDict[MQTTRPCClient.ARGS_DICT_KEY])
        else:
            return ""

    def rpcCall(self, methodName, argList):
        """@brief Call an RPC
           @param methodName The name of the RPC to call. Currently this is just the method name.
                             This method could exist on an instance of any object. We could have
                             made this include instance details but to keep things simple
                             just the method name is provided.
           @param argList The arguments for the RPC
           @return The response to the RPC."""
        self._sendMsgDict={}
        self._sendMsgDict[MQTTRPCClient.CLIENT_ID_DICT_KEY]=self._clientSRC
        self._sendMsgDict[MQTTRPCClient.METHOD_DICT_KEY]=methodName
        self._sendMsgDict[MQTTRPCClient.ARGS_DICT_KEY]=argList

        jsonDict = json.dumps(self._sendMsgDict, indent=4, sort_keys=True)

        self._client.subscribe(self._clientSRC)

        self._condition.acquire()

        response = None
        try:
            self._client.publish(self._serverRPCTopic, jsonDict)

            response = self._getResponse()

        finally:

            self._condition.release()

        return response



#!/usr/bin/env python3

import sys
import os
import re
import traceback
from   threading import Lock
from   socket import socket, AF_INET, SOCK_DGRAM
from   getpass import getpass, getuser
from   time import strftime, localtime
from   datetime import datetime
        
class UIO(object):
    """@brief responsible for user output and input via stdout/stdin"""

    DISPLAY_ATTR_RESET          =   0
    DISPLAY_ATTR_BRIGHT         =   1
    DISPLAY_ATTR_DIM            =   2
    DISPLAY_ATTR_UNDERSCORE     =   4
    DISPLAY_ATTR_BLINK          =   5
    DISPLAY_ATTR_REVERSE        =   7
    DISPLAY_ATTR_HIDDEN         =   8

    DISPLAY_ATTR_FG_BLACK       =   30
    DISPLAY_ATTR_FG_RED         =   31
    DISPLAY_ATTR_FG_GREEN       =   32
    DISPLAY_ATTR_FG_YELLOW      =   33
    DISPLAY_ATTR_FG_BLUE        =   34
    DISPLAY_ATTR_FG_MAGNETA     =   35
    DISPLAY_ATTR_FG_CYAN        =   36
    DISPLAY_ATTR_FG_WHITE       =   37

    DISPLAY_ATTR_BG_BLACK       =   40
    DISPLAY_ATTR_BG_RED         =   41
    DISPLAY_ATTR_BG_GREEN       =   42
    DISPLAY_ATTR_BG_YELLOW      =   43
    DISPLAY_ATTR_BG_BLUE        =   44
    DISPLAY_ATTR_BG_MAGNETA     =   45
    DISPLAY_ATTR_BG_CYAN        =   46
    DISPLAY_ATTR_BG_WHITE       =   47

    DISPLAY_RESET_ESCAPE_SEQ    = "\x1b[0m"

    PROG_BAR_LENGTH             =   40

    USER_LOG_SYM_LINK           = "log.txt"
    DEBUG_LOG_SYM_LINK          = "debug_log.txt"
  
    @staticmethod
    def GetInfoEscapeSeq():
        """@return the info level ANSI escape sequence."""
        return "\x1b[{:01d};{:02d}m".format(UIO.DISPLAY_ATTR_FG_GREEN, UIO.DISPLAY_ATTR_BRIGHT)

    @staticmethod
    def GetDebugEscapeSeq():
        """@return the debug level ANSI escape sequence."""
        return "\x1b[{:01d};{:02d};{:02d}m".format(UIO.DISPLAY_ATTR_FG_BLACK, UIO.DISPLAY_ATTR_BG_WHITE, UIO.DISPLAY_ATTR_BRIGHT)

    @staticmethod
    def GetWarnEscapeSeq():
        """@return the warning level ANSI escape sequence."""
        return "\x1b[{:01d};{:02d}m".format(UIO.DISPLAY_ATTR_FG_RED, UIO.DISPLAY_ATTR_BRIGHT)

    @staticmethod
    def GetErrorEscapeSeq():
        """@return the warning level ANSI escape sequence."""
        return "\x1b[{:01d};{:02d}m".format(UIO.DISPLAY_ATTR_FG_RED, UIO.DISPLAY_ATTR_BLINK)

    @staticmethod
    def RemoveEscapeSeq(text):
        """@brief Remove ANSI escape sequences that maybe present in text.
           @param text A string that may contain ANSI escape sequences.
           @return The text with any ANSI escape sequences removed."""
        escapeSeq =re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
        return escapeSeq.sub('', text)

    def __init__(self, debug=False, colour=True):
        self._debug                         = debug
        self._colour                        = colour
        self._logFile                       = None
        self._progBarSize                   = 0
        self._progBarGrow                   = True
        self._debugLogEnabled               = False
        self._debugLogFile                  = None
        self._symLinkDir                    = None
        self._sysLogEnabled                 = False
        self._sysLogHost                    = None

    def logAll(self, enabled):
        """@brief Turn on/off the logging of all output including debug output even if debugging is off."""
        self._debugLogEnabled = enabled

    def enableDebug(self, enabled):
        """@brief Enable/Disable debugging
           @param enabled If True then debugging is enabled"""
        self._debug = enabled

    def isDebugEnabled(self):
        """@return True if debuggin is eenabled."""
        return self._debug

    def info(self, text, highlight=False):
        """@brief Present an info level message to the user.
           @param text The line of text to be presented to the user."""
        if self._colour:
            if highlight:
                self._print('{}INFO:  {}{}'.format(UIO.GetInfoEscapeSeq(), text, UIO.DISPLAY_RESET_ESCAPE_SEQ))
            else:
                self._print('{}INFO{}:  {}'.format(UIO.GetInfoEscapeSeq(), UIO.DISPLAY_RESET_ESCAPE_SEQ, text))
        else:
            self._print('INFO:  {}'.format(text))
        self._update_syslog(PRIORITY.INFO, text)

    def debug(self, text):
        """@brief Present a debug level message to the user if debuging is enabled.
           @param text The line of text to be presented to the user."""
        if self._debug:
            if self._colour:
                self._print('{}DEBUG{}: {}'.format(UIO.GetDebugEscapeSeq(), UIO.DISPLAY_RESET_ESCAPE_SEQ, text))
            else:
                self._print('DEBUG: {}'.format(text))
        elif self._debugLogEnabled and self._debugLogFile:
            if self._colour:
                self.storeToDebugLog('{}DEBUG{}: {}'.format(UIO.GetDebugEscapeSeq(), UIO.DISPLAY_RESET_ESCAPE_SEQ, text))
            else:
                self.storeToDebugLog('DEBUG: {}'.format(text))
        self._update_syslog(PRIORITY.DEBUG, text)
        
    def warn(self, text):
        """@brief Present a warning level message to the user.
           @param text The line of text to be presented to the user."""
        if self._colour:
            self._print('{}WARN{}:  {}'.format(UIO.GetWarnEscapeSeq(), UIO.DISPLAY_RESET_ESCAPE_SEQ, text))
        else:
            self._print('WARN:  {}'.format(text))
        self._update_syslog(PRIORITY.WARNING, text)
        
    def error(self, text):
        """@brief Present an error level message to the user.
           @param text The line of text to be presented to the user."""
        if self._colour:
            self._print('{}ERROR{}: {}'.format(UIO.GetErrorEscapeSeq(), UIO.DISPLAY_RESET_ESCAPE_SEQ, text))
        else:
            self._print('ERROR: {}'.format(text))
        self._update_syslog(PRIORITY.ERROR, text)

    def _print(self, text):
        """@brief Print text to stdout"""
        self.storeToLog(text)
        if self._debugLogEnabled and self._debugLogFile:
            self.storeToDebugLog(text)
        print(text)
      
    def getInput(self, prompt, noEcho=False, stripEOL=True):
        """@brief Get a line of text from the user.
           @param noEcho If True then * are printed when each character is pressed.
           @param stripEOL If True then all end of line (\r, \n) characters are stripped.
           @return The line of text entered by the user."""
        if self._colour:
            if noEcho:
                prompt = "{}INPUT{}: ".format(UIO.GetInfoEscapeSeq(), UIO.DISPLAY_RESET_ESCAPE_SEQ) + prompt + ": "
                self.storeToLog(prompt, False)
                response = getpass(prompt, sys.stdout)

            else:
                prompt = "{}INPUT{}: ".format(UIO.GetInfoEscapeSeq(), UIO.DISPLAY_RESET_ESCAPE_SEQ) + prompt + ": "
                self.storeToLog(prompt, False)
                response = input(prompt)

        else:
            if noEcho:
                prompt = "INPUT: " + prompt + ": "
                self.storeToLog(prompt, False)
                response = getpass(prompt, sys.stdout)

            else:
                prompt = "INPUT: " + prompt + ": "
                self.storeToLog(prompt, False)
                response = input(prompt)

        if stripEOL:
            response = response.rstrip('\n')
            response = response.rstrip('\r')

        self.storeToLog(response)
        return response

    def getBoolInput(self, prompt, allowQuit=True):
        """@brief Get boolean repsonse from user (y or n response).
           @param allowQuit If True and the user enters q then the program will exit.
           @return True or False"""
        while True:
            response = self.getInput(prompt=prompt)
            if response.lower() == 'y':
                return True
            elif response.lower() == 'n':
                return False
            elif allowQuit and response.lower() == 'q':
                sys.exit(0)

    def getIntInput(self, prompt, allowQuit=True, radix=10):
      """@brief Get a decimal int number from the user.
         @param allowQuit If True and the user enters q then the program will exit.
         @param radix The radix of the number entered (default=10).
         @return True or False"""
      while True:
        response = self.getInput(prompt=prompt)
        try:
          return int(response, radix)
        except ValueError:
          self.warn("%s is not a valid integer value." % (response))

        if allowQuit and response.lower() == 'q':
          return None

    def errorException(self):
        """@brief Show an exception traceback if debugging is enabled"""
        if self._debug:
            lines = traceback.format_exc().split('\n')
            for l in lines:
                self.error(l)

    def getPassword(self, prompt):
        """@brief Get a password from a user.
           @param prompt The user prompt.
           @return The password entered."""
        return self.getInput(prompt, noEcho=True)

    def setSymLinkDir(self, symLinkDir):
        """@brief Set a shortcut location for symLink.
           @param symLinkDir The directory to create the simlink.
           @return None"""
        self._symLinkDir=symLinkDir

    def setLogFile(self, logFile):
        """@brief Set a logfile for all output.
           @param logFile The file to send all output to.
           @return None"""
        self._logFile=logFile
        self._debugLogFile = "{}.debug.txt".format(self._logFile)

    def storeToLog(self, text, addLF=True, addDateTime=True):
        """@brief Save the text to the main log file if one is defined.
           @param text The text to be saved.
           @param addLF If True then a line feed is added to the output in the log file.
           @return None"""
        self._storeToLog(text, self._logFile, addLF=addLF, addDateTime=addDateTime)

    def storeToDebugLog(self, text, addLF=True, addDateTime=True):
        """@brief Save the text to the debug log file if one is defined. This file holds all the
                  data from the main log file plus debug data even if debugging is not enabled.
           @param text The text to be saved.
           @param addLF If True then a line feed is added to the output in the log file.
           @return None"""
        self._storeToLog(text, self._debugLogFile, addLF=addLF, addDateTime=addDateTime, symLinkFile=UIO.DEBUG_LOG_SYM_LINK)

    def _storeToLog(self, text, logFile, addLF=True, addDateTime=True, symLinkFile=USER_LOG_SYM_LINK):
        """@brief Save the text to the log file if one is defined.
           @param text The text to be saved.
           @param logFile The logFile to save data to.
           @param addLF If True then a line feed is added to the output in the log file.
           @param symLinkFile The name of the fixed symlink file to point to the latest log file.
           @return None"""
        createSymLink = False
        if logFile:
            if addDateTime:
                timeStr = datetime.now().strftime("%d/%m/%Y-%H:%M:%S.%f")
                text = "{}: {}".format(strftime(timeStr, localtime()).lower(), text)

            # If the log file is about to be created then we will create a symlink
            # to the file.
            if not os.path.isfile(logFile):
                createSymLink = True

            fd = open(logFile, 'a')
            if addLF:
                fd.write("{}\n".format(text))
            else:
                fd.write(text)
            fd.close()

            if createSymLink:
                #This is helpful as the link will point to the latest log file
                #which can be useful when debugging. I.E  no need to find the
                #name of the latest file.
                dirName = self._symLinkDir
                # if the simlink has not been set then default to the logging file
                if dirName is None:
                    dirName = os.path.dirname(logFile)
                absSymLink = os.path.join(dirName, symLinkFile)
                if os.path.lexists(absSymLink):
                    os.remove(absSymLink)
                os.symlink(logFile, absSymLink)

    def showProgBar(self, barChar='*'):
        """@brief Show a bar that grows and shrinks to indicate an activity is occuring."""
        if self._progBarGrow:
            sys.stdout.write(barChar)
            self._progBarSize+=1
            if self._progBarSize > UIO.PROG_BAR_LENGTH:
                self._progBarGrow=False
        else:
            sys.stdout.write('\b')
            sys.stdout.write(' ')
            sys.stdout.write('\b')
            self._progBarSize-=1
            if self._progBarSize == 0:
                self._progBarGrow=True

        sys.stdout.flush()

    def clearProgBar(self):
        """@brief Clear any progress characters that maybe present"""
        sys.stdout.write('\b' * UIO.PROG_BAR_LENGTH)
        sys.stdout.write(' '*UIO.PROG_BAR_LENGTH)
        sys.stdout.write('\r')
        sys.stdout.flush()

    def getLogFile(self):
        """@return the name of the user output log file or None if not set"""
        return self._logFile

    def getDebugLogFile(self):
        """@return The name of the debug log file or None if not set."""
        return self._debugLogFile

    def getCurrentUsername(self):
        """Get the current users username or return unknown_user if not able to read it.
           This is required as getpass.getuser() does not always work on windows platforms."""
        username="unknown_user"
        try:
            username=getuser()
        except:
            pass
        return username

    def enableSyslog(self, enabled, host="localhost"):
        """@brief Enable/disable syslog.
           @param enabled If True then syslog is enabled."""
        self._sysLogEnabled = enabled
        self._sysLogHost = host
      
    def _update_syslog(self, pri, msg):
        """Send a message to syslog is syslog is enabled
        Syslog messages will have the following components
        0 = time/date stamp
        1 = hostname
        2 = main python file name
        3 = PID
        4 = username under which the program is being executed
        5 = The syslog message

        The syslog messages will be prefixed withj the application name
        """
        if self._sysLogEnabled:
            aMsg=msg
            #Ensure we have no zero characters in the message. 
            # syslog will throw an error if it finds any
            if "\x00" in aMsg:
                aMsg=aMsg.replace("\x00", "")

            try:
                #send aMsg to syslog with the current process ID and username
                syslog(pri, "%d %s: %s" % (os.getpid(), str(self.getCurrentUsername()), str(aMsg) ), host=self._sysLogHost )
            #Ignore if se can't resolve address. We don't really syslog want errors to stop the user interface
            except:
                pass

class ConsoleMenu(object):
    """@brief Responsible for presenting a list of options to the user on a
              console/terminal interface and allowing the user to select 
              the options as required."""
    def __init__(self, uio):
        """@brief Constructor
           @param uio A UIO instance."""
        self._uio = uio
        self._menuList = []
        
    def add(self, menuStr, menuMethod, args=None):
        """@brief Add a menu option.
           @param menuStr The String displayed as the menu option.
           @param menuMethod The method to be called when this option is selected.
           @param args An optional list or tuple of arguments to pass to the method."""
        self._menuList.append( (menuStr, menuMethod, args) )
        
    def show(self, showSelectedOption=False):
        """@brief Show the menu to the user and allow the user to interact with it.
           @param showSelectedOption If True then the option selected is displayed before executing the method."""
        while True:
            selectorId = 1
            for menuStr, _, _ in self._menuList:
                self._uio.info("{: >2}: {}".format(selectorId, menuStr))
                selectorId = selectorId + 1
            selectedID = self._uio.getIntInput("Select a menu option")
            if selectedID >= 1 and selectedID <= len(self._menuList):
                menuStr, selectedMethod, args = self._menuList[selectedID-1]
                if showSelectedOption:
                    self._uio.info(menuStr)
                if not args:
                    selectedMethod()
                else:
                    selectedMethod(*args)

# -------------------------------------------------------------------
# @brief An implementation to allow syslog messages to be generated.
# 
class FACILITY:
  KERN=0
  USER=1
  MAIL=2
  DAEMON=3
  AUTH=4
  SYSLOG=5
  LPR=6
  NEWS=7
  UUCP=8
  CRON=9
  AUTHPRIV=10
  FTP=11
  LOCAL0=16
  LOCAL1=17
  LOCAL2=18
  LOCAL3=19
  LOCAL4=20
  LOCAL5=21
  LOCAL6=22
  LOCAL7=23
 
class PRIORITY:
  EMERG=0
  ALERT=1
  CRIT=2
  ERROR=3
  WARNING=4
  NOTICE=5
  INFO=6
  DEBUG=7

syslogSocket=None
lock = Lock()
  
def syslog(priority, message, facility=FACILITY.LOCAL0, host='localhost', port=514):
    """
	  @brief Send a syslog message.
      @param priority The syslog priority level.
      @param message The text message to be sent.
      @param facility The syslog facility 
      @param host The host address for the systlog server.
      @param port The syslog port.
	"""
    global lock, timer, syslogSocket
	
    try:
        lock.acquire()
        if not syslogSocket:
            syslogSocket = socket(AF_INET, SOCK_DGRAM)

        smsg = '<%05d>%s' % ( (priority + facility*8), message )
        syslogSocket.sendto(smsg.encode('ascii', 'ignore'), (host, port))

    finally:
        lock.release()



#!/usr/bin/env python3

import os
import plotly
import tarfile
import tempfile
import shutil

from   pathlib import Path
try:
    from   .ssh import SSH
    from   .helper import getHomePath
except:
    #We get here if running netplotly_demo.py as it sits in the same folder
    #as netplotly.py
    from   ssh import SSH
    from   helper import getHomePath

class NetPlotly(object):
    """@brief Manage plotly graphs on a webserver and allow them to be transferred over a network via ssh.
              Plots are saved to a local folder. The local folder maybe served via a web server.
              The contents of the local folder maybe transferred to a remote machine via ssh.
              The contents of the remote folder maybe served via a web server.

              The local and remote folders will contain an index.html file. This displays a table of
              all the available plots. Each plot name in the table maybe clicked which will display
              the plotly plot."""
    INDEX_HTML_FILE         = "netplotly_index.html"
    DEFAULT_LOCAL_ROOT      = "~/.netplotly"
    ASSETS_FOLDER           = "assets"
    PLOT_LIST_FILE          = "plot_list.txt"

    @staticmethod
    def GetTarBallFile():
        """@brief Get the tarball file path
           @return The abs file path"""
        return os.path.join( tempfile.gettempdir(), "netplotly.tgz" )

    def __init__(self, localRoot=DEFAULT_LOCAL_ROOT, serverRoot=DEFAULT_LOCAL_ROOT, host=None, username=None, password=None, port=22, uio=None):
        """@brief Constructor
           @param localRoot The local folder to store the plots in.
                  A web server may serve files directly from here or they maybe
                  uploaded (via ssh) to a webserver using the API's provided
                  by this class.
           @param serverRoot The folder to store html files in on the server.
           @param host The address of the ssh server to upload files to.
           @param username The SSH server username
           @param password The SSH server password. Not required if auto login via public
                           key is enabled for the ssh server.
           @param port The ssh server port (default=22).
           @param uio A UIO instance for displaying messages to the user."""
        self._localRoot = localRoot
        self._serverRoot = serverRoot
        self._host       = host
        self._username  = username
        self._password  = password
        self._port      = port
        self._uio       = uio
        self._ssh       = None

        self._updateLocalRoot()

    def _updateLocalRoot(self):
        """@brief Set the local _rootp attr"""
        if self._localRoot.startswith("~"):
            self._localRoot = getHomePath() + self._localRoot.replace("~", "")

    def info(self, msg):
        """@brief Display an info level message if we have been provided with a uio object"""
        if self._uio:
            self._uio.info(msg)

    def _getPlotListFile(self):
        """@return The plot list file"""
        return os.path.join(self._localRoot, NetPlotly.PLOT_LIST_FILE)

    def _updatePlotList(self, plotTitle):
        """@brief Update the plot title log file. This file is read by the javascript to
                  present a table of available plots to the user. The PLOT_LIST_FILE
                  simply contains a number of lines each lines text is the name of the plot.
           @param plotTitle The plot title
           @return None"""
        alreadyAdded = False
        plotListFile = self._getPlotListFile()
        if os.path.isfile(plotListFile):
            fd = open(plotListFile, 'r')
            lines = fd.readlines()
            fd.close()
            for line in lines:
                line=line.rstrip("\n")
                if line == plotTitle:
                    alreadyAdded=True
        if not alreadyAdded:
            fd = open(plotListFile, 'a')
            fd.write("{}\n".format(plotTitle))
            fd.close()

    def _removePlotList(self, plotTitle):
        """@brief Remove a file from the plot list.
           @param plotTitle The plot title
           @return None"""
        plotListFile = self._getPlotListFile()
        newLines = []
        if os.path.isfile(plotListFile):
            fd = open(plotListFile, 'r')
            lines = fd.readlines()
            fd.close()

            for line in lines:
                if not line.startswith(plotTitle):
                    newLines.append(line)

            fd = open(plotListFile, 'w')
            for l in newLines:
                fd.write(l)

            fd.close()

    def save(self, fig, htmlFile=None, autoOpen=False):
        """@brief Save a plotly figure to an html file in the local home folder.
           @param fig The plotly figure
           @param htmlFile The name of the htmlfile to save. If not provided the name of the plot
                           with the .html suffix is used as the filename.
           @param autoOpen If True then a browser window is launched to show the plot"""
        #don't allow index.html as this is usd for the table of plots
        if htmlFile == NetPlotly.INDEX_HTML_FILE:
            raise Exception("save(): {} is a reserved html file name.".format(NetPlotly.INDEX_HTML_FILE))

        #If no html file provided then use the name of the plot
        if not htmlFile:
            htmlFile = "{}.html".format( fig["layout"]["title"]["text"] )

        if not os.path.isdir(self._localRoot):
            os.makedirs(self._localRoot)
            self.info("Created {}".format(self._localRoot))

        plotName = htmlFile.replace(".html", "")
        self._updatePlotList(plotName)

        fileToSave = os.path.join(self._localRoot, htmlFile)
        plotly.offline.plot(fig, filename=fileToSave, auto_open = autoOpen)
        self.info("Saved {}".format(fileToSave))

    def _createTarBall(self, opFile, srcFolder):
        self.info("Creating {} from {}".format(opFile, srcFolder))
        with tarfile.open(opFile, "w:gz") as tar:
            tar.add(srcFolder, arcname=os.path.basename(srcFolder))

    def connect(self):
        """@brief connect to the server vis ssh"""
        if not self._host:
            raise Exception("No SSH server defined to connect to")

        if not self._ssh:
            self.info("Connecting to {}:{} as {}".format(self._host, self._port, self._username))
            self._ssh = SSH(self._host, username=self._username, password=self._password, port=self._port)
            self._ssh.connect(connectSFTPSession=True)
            self.info("Connected")

    def disconnect(self):
        """@brief Close the connection to the server."""
        if self._ssh:
            self._ssh.close()
            self._ssh = None
            self.info("Closed server connection")

    def upload(self, purge=False):
        """@brief Upload the plot to the server.
           @param purge If True (default is False) then all html files are removed from the server before uploading
                        from the local path. If True the caller must be careful to set serverRoot (in Constructor)
                        as all files in this location are purged before files are uploaded."""
        if self._host and self._username:
            tarBallFile = NetPlotly.GetTarBallFile()
            self._createTarBall(tarBallFile, self._localRoot)
            self.connect()

            if purge:
                #Remove all the files in the server folder
                cmd = "rm -f {}/*".format(self._serverRoot)
                self._ssh.runCmd(cmd)

            self.info("Uploading {} to {}:{}".format(tarBallFile, self._host, self._port))
            self._ssh.putFile(tarBallFile, tarBallFile)

            self.info("Decompressing {} into {} on the server.".format(tarBallFile, self._serverRoot))
            rPath = Path(self._serverRoot)
            cmd = "tar zxvf {} -C {}".format(tarBallFile, rPath.parent)
            self._ssh.runCmd(cmd)

            self.disconnect()

    def getPlotNameList(self):
        """@brief Get a list of the names of all the plots currently stored in the local root folder.
           @return a list of plot names."""
        plotList = []
        plotListFile = self._getPlotListFile()
        if os.path.isfile(plotListFile):
            fd = open(plotListFile, 'r')
            plotList = fd.read().splitlines()
            fd.close()

        return plotList

    def getPlotFileList(self):
        """@brief Get a list of the plot file names of all the plots currently stored in the local root folder.
           @return a list of plot file names."""
        plotList = self.getPlotNameList()
        plotFileList = ["{}.html".format(plotName) for plotName in plotList]
        return plotFileList

    def remove(self, plotName):
        """@brief Remove a plot by it's name from the plot list.
           @return None"""
        fileToRemove = os.path.join(self._localRoot, "{}.html".format(plotName))
        if os.path.isfile(fileToRemove):
            os.remove(fileToRemove)
        self._removePlotList(plotName)

    def removeLocalRoot(self):
        """@brief Remove all plots from the list of plots. Be sure you have _localRoot set
                  correctly before calling this method.
           @return None"""
        shutil.rmtree(self._localRoot)



#!/bin/sh/env python3d

import  os
import  base64
import  datetime
import  sys
import  json
import  copy

from    shutil import copyfile

from    os.path import join, expanduser, getmtime, basename, isdir, isfile
from    .helper import getHomePath

class ConfigManager(object):
    """@brief Responsible for storing and loading configuration.
              Also responsible for providing methods that allow users to enter
              configuration."""

    UNSET_VALUE                 = "UNSET"
    DECIMAL_INT_NUMBER_TYPE     = 0
    HEXADECIMAL_INT_NUMBER_TYPE = 1
    FLOAT_NUMBER_TYPE           = 2
    SSH_FOLDER                  = ".ssh"
    PRIVATE_SSH_KEY_FILENAME    = "id_rsa"

    @staticmethod
    def GetString(uio, prompt, previousValue, allowEmpty=True):
      """@brief              Get a string from the the user.
         @param uio          A UIO (User Inpu Output) instance.
         @param prompt       The prompt presented to the user in order to enter
                             the float value.
         @param previousValue The previous value of the string.
         @param allowEmpty   If True then allow the string to be empty."""
      _prompt = prompt
      try:
          prompt = "%s (%s)" % (prompt, previousValue)
      except ValueError:
          prompt = "%s" % (prompt)

      while True:

        response = uio.getInput("%s" % (prompt))

        if len(response) == 0:

            if allowEmpty:

                if len(previousValue) > 0:
                    booleanResponse = uio.getInput("Do you wish to enter the previous value '%s' y/n: " % (previousValue) )
                    booleanResponse=booleanResponse.lower()
                    if booleanResponse == 'y':
                        response = previousValue
                        break

                booleanResponse = uio.getInput("Do you wish to clear '%s' y/n: " % (_prompt) )
                booleanResponse=booleanResponse.lower()
                if booleanResponse == 'y':
                    break
                else:
                    uio.info("A value is required. Please enter a value.")

            else:

                booleanResponse = uio.getInput("Do you wish to enter the previous value of %s y/n: " % (previousValue) )
                booleanResponse=booleanResponse.lower()
                if booleanResponse == 'y':
                    response = previousValue
                    break
        else:
            break

      return response

    @staticmethod
    def IsValidDate(date):
        """@brief determine if the string is a valid date.
           @param date in the form DAY/MONTH/YEAR (02:01:2018)"""
        validDate = False
        if len(date) >= 8:
            elems = date.split("/")
            try:
                day = int(elems[0])
                month = int(elems[1])
                year = int(elems[2])
                datetime.date(year, month, day)
                validDate = True
            except ValueError:
                pass
        return validDate

    @staticmethod
    def GetDate(uio, prompt, previousValue, allowEmpty=True):
        """@brief Input a date in the format DAY:MONTH:YEAR"""
        if not ConfigManager.IsValidDate(previousValue):
            today =  datetime.date.today()
            previousValue = today.strftime("%d/%m/%Y")
        while True:
            newValue = ConfigManager.GetString(uio, prompt, previousValue, allowEmpty=allowEmpty)
            if ConfigManager.IsValidDate(newValue):
                return newValue

    @staticmethod
    def IsValidTime(theTime):
        """@brief determine if the string is a valid time.
           @param theTime in the form HOUR:MINUTE:SECOND (12:56:01)"""
        validTime = False
        if len(theTime) >= 5:
            elems = theTime.split(":")
            try:
                hour = int(elems[0])
                minute = int(elems[1])
                second = int(elems[2])
                datetime.time(hour, minute, second)
                validTime = True
            except ValueError:
                pass
        return validTime

    @staticmethod
    def GetTime(uio, prompt, previousValue, allowEmpty=True):
        """@brief Input a time in the format HOUR:MINUTE:SECOND"""
        if not ConfigManager.IsValidTime(previousValue):
            today =  datetime.datetime.now()
            previousValue = today.strftime("%H:%M:%S")
        while True:
            newValue = ConfigManager.GetString(uio, prompt, previousValue, allowEmpty=allowEmpty)
            if ConfigManager.IsValidTime(newValue):
                return newValue

    @staticmethod
    def _GetNumber(uio, prompt, previousValue=UNSET_VALUE, minValue=UNSET_VALUE, maxValue=UNSET_VALUE, numberType=FLOAT_NUMBER_TYPE, radix=10):
      """@brief              Get float repsonse from user.
         @param uio          A UIO (User Inpu Output) instance.
         @param prompt       The prompt presented to the user in order to enter
                             the float value.
         @param previousValue The previous number value.
         @param minValue     The minimum acceptable value.
         @param maxValue     The maximum acceptable value.
         @param numberType   The type of number."""

      if numberType == ConfigManager.DECIMAL_INT_NUMBER_TYPE:

          radix=10

      elif numberType == ConfigManager.HEXADECIMAL_INT_NUMBER_TYPE:

          radix = 16

      while True:

        response = ConfigManager.GetString(uio, prompt, previousValue, allowEmpty=False)

        try:

            if numberType == ConfigManager.FLOAT_NUMBER_TYPE:

                value = float(response)

            else:

                value = int(str(response), radix)

            if minValue != ConfigManager.UNSET_VALUE and value < minValue:

                if radix == 16:
                    minValueStr = "0x%x" % (minValue)
                else:
                    minValueStr = "%d" % (minValue)

                uio.info("%s is less than the min value of %s." % (response, minValueStr) )
                continue

            if maxValue != ConfigManager.UNSET_VALUE and value > maxValue:

                if radix == 16:
                    maxValueStr = "0x%x" % (maxValue)
                else:
                    maxValueStr = "%d" % (maxValue)

                uio.info("%s is greater than the max value of %s." % (response, maxValueStr) )
                continue

            return value

        except ValueError:

            if numberType == ConfigManager.FLOAT_NUMBER_TYPE:

                uio.info("%s is not a valid number." % (response) )

            elif numberType == ConfigManager.DECIMAL_INT_NUMBER_TYPE:

                uio.info("%s is not a valid integer value." % (response) )

            elif numberType == ConfigManager.HEXADECIMAL_INT_NUMBER_TYPE:

                uio.info("%s is not a valid hexadecimal value." % (response) )

    @staticmethod
    def GetDecInt(uio, prompt, previousValue=UNSET_VALUE, minValue=UNSET_VALUE, maxValue=UNSET_VALUE):
      """@brief              Get a decimal integer number from the user.
         @param uio          A UIO (User Inpu Output) instance.
         @param prompt       The prompt presented to the user in order to enter
                             the float value.
         @param previousValue The previous number value.
         @param minValue     The minimum acceptable value.
         @param maxValue     The maximum acceptable value."""

      return ConfigManager._GetNumber(uio, prompt, previousValue=previousValue, minValue=minValue, maxValue=maxValue, numberType=ConfigManager.DECIMAL_INT_NUMBER_TYPE)

    @staticmethod
    def GetHexInt(uio, prompt, previousValue=UNSET_VALUE, minValue=UNSET_VALUE, maxValue=UNSET_VALUE):
      """@brief              Get a decimal integer number from the user.
         @param uio          A UIO (User Inpu Output) instance.
         @param prompt       The prompt presented to the user in order to enter
                             the float value.
         @param previousValue The previous number value.
         @param minValue     The minimum acceptable value.
         @param maxValue     The maximum acceptable value."""

      return ConfigManager._GetNumber(uio, prompt, previousValue=previousValue, minValue=minValue, maxValue=maxValue, numberType=ConfigManager.HEXADECIMAL_INT_NUMBER_TYPE)

    @staticmethod
    def GetFloat(uio, prompt, previousValue=UNSET_VALUE, minValue=UNSET_VALUE, maxValue=UNSET_VALUE):
      """@brief              Get a float number from the user.
         @param uio          A UIO (User Inpu Output) instance.
         @param prompt       The prompt presented to the user in order to enter
                             the float value.
         @param previousValue The previous number value.
         @param minValue     The minimum acceptable value.
         @param maxValue     The maximum acceptable value."""

      return ConfigManager._GetNumber(uio, prompt, previousValue, minValue=minValue, maxValue=maxValue, numberType=ConfigManager.FLOAT_NUMBER_TYPE)

    @staticmethod
    def GetPrivateKeyFile():
        """@brief Get the private key file."""
        homePath = getHomePath()
        folder = os.path.join(homePath, ConfigManager.SSH_FOLDER)
        priKeyFile = os.path.join(folder, ConfigManager.PRIVATE_SSH_KEY_FILENAME)
        if not os.path.isfile(priKeyFile):
            raise Exception("%s file not found" % (priKeyFile) )
        return priKeyFile

    @staticmethod
    def GetPrivateSSHKeyFileContents():
        """@brief Get the private ssh key file.
           @return The key file contents"""
        privateRSAKeyFile = ConfigManager.GetPrivateKeyFile()
        fd = open(privateRSAKeyFile, 'r')
        fileContents = fd.read()
        fd.close()
        return fileContents

    @staticmethod
    def GetCrypter():
        """@brief Get the object responsible for encrypting and decrypting strings."""
        # Only import the cryptography module if it is used so as to avoid the
        # necessity of having the  cryptography module to use the pconfig module.
        from    cryptography.fernet import Fernet
        keyString = ConfigManager.GetPrivateSSHKeyFileContents()
        keyString = keyString[60:92]
        priKeyBytes = bytes(keyString, 'utf-8')
        priKeyBytesB64 = base64.b64encode(priKeyBytes)
        return Fernet(priKeyBytesB64)

    @staticmethod
    def Encrypt(inputString):
        """@brief Encrypt the string.
           @param inputString The string to encrypt.
           @return The encrypted string"""
        crypter = ConfigManager.GetCrypter()
        token = crypter.encrypt(bytes(inputString, 'utf-8'))
        return token.decode('utf-8')

    @staticmethod
    def Decrypt(inputString):
        """@brief Decrypt the string.
           @param inputString  The string to decrypt.
           return The decrypted string"""
        crypter = ConfigManager.GetCrypter()
        token = inputString.encode('utf-8')
        decryptedBytes = crypter.decrypt(token)
        return decryptedBytes.decode('utf-8')

    def __init__(self, uio, cfgFilename, defaultConfig, addDotToFilename=True, encrypt=False, cfgPath=None):
        """@brief Constructor
           @param uio A UIO (User Inpu Output) instance.
           @param cfgFilename   The name of the config file.
           @param defaultConfig A default config instance containg all the default key-value pairs.
           @param addDotToFilename If True (default) then a . is added to the start of the filename. This hides the file in normal cases.
           @param encrypt If True then data will be encrypted in the saved files.
                          The encryption uses part of the the local SSH RSA private key.
                          This is not secure but assuming the private key has not been compromised it's
                          probably the best we can do. Therefore if encrypt is set True then the
                          an ssh key must be present in the ~/.ssh folder named id_rsa.
           @param cfgPath The config path when the config file will be stored. By default this is unset and the 
                          current users home folder is the location of the config file."""
        self._uio               = uio
        self._cfgFilename       = cfgFilename
        self._defaultConfig     = defaultConfig
        self._addDotToFilename  = addDotToFilename
        self._encrypt           = encrypt
        self._cfgPath           = cfgPath
        self._configDict        = {}

        self._cfgFile = self._getConfigFile()
        self._modifiedTime = self._getModifiedTime()

    def _getConfigFile(self):
        """@brief Get the config file."""

        if not self._cfgFilename:
            raise Exception("No config filename defined.")

        cfgFilename = self._cfgFilename
        if self._addDotToFilename:
            if not self._cfgFilename.startswith("."):

                cfgFilename=".%s" % (self._cfgFilename)

            else:

                cfgFilename=self._cfgFilename

        #The the config path has been set then use it
        if self._cfgPath:
            configPath = self._cfgPath

        else:
            configPath=""
            #If an absolute path is set for the config file then don't try to 
            #put the file in the users home dir
            if not self._cfgFilename.startswith("/"):
                configPath = expanduser("~")
                configPath = configPath.strip()
                #If no user is known then default to root user.
                #This occurs on Omega2 startup apps.
                if len(configPath) == 0 or configPath == '/':
                    configPath="/root"

        return join( configPath, cfgFilename )

    def addAttr(self, key, value):
        """@brief Add an attribute value to the config.
           @param key The key to store the value against.
           @param value The value to be stored."""
        self._configDict[key]=value

    def getAttrList(self):
        """@return A list of attribute names that are stored."""
        return self._configDict.keys()

    def getAttr(self, key, allowModify=True):
        """@brief Get an attribute value.
           @param key The key for the value we're after.
           @param allowModify If True and the configuration has been modified
                  since the last read by the caller then the config will be reloaded."""

        #If the config file has been modified then read the config to get the updated state.
        if allowModify and self.isModified():
            self.load(showLoadedMsg=False)
            self.updateModifiedTime()

        return self._configDict[key]

    def _saveDict(self, dictToSave):
      """@brief Save dict to a file.
         @param dictToSave The dictionary to save."""

      try:

          if self._encrypt:
              stringToSave = json.dumps(dictToSave, sort_keys=True)
              stringToSave = ConfigManager.Encrypt(stringToSave)
              fd = open(self._cfgFile, "w")
              fd.write(stringToSave)
              fd.close()
          else:
              json.dump(dictToSave, open(self._cfgFile, "w"), sort_keys=True)

          if self._uio:
              self._uio.info("Saved config to %s" % (self._cfgFile) )

      except IOError as i:
        raise IOError(i.errno, 'Failed to write file \'%s\': %s'
            % (self._cfgFile, i.strerror), i.filename).with_traceback(i)

    def store(self, copyToRoot=False):
        """@brief Store the config to the config file.
           @param copyToRoot If True copy the config to the root user (Linux only)
                             if not running with root user config path. If True
                             on non Linux system config will only be saved in
                             the users home path. Default = False."""
        self._saveDict(self._configDict)

        self.updateModifiedTime()

        if copyToRoot and not self._cfgFile.startswith("/root/") and isdir("/root"):
            fileN = basename(self._cfgFile)
            rootCfgFile = join("/root", fileN)
            copyfile(self._cfgFile, rootCfgFile)
            if self._uio:
                self._uio.info("Also updated service list in %s" % (rootCfgFile))

    def _getDict(self):
      """@brief Load dict from file
         @return Return the dict loaded from the file."""
      dictLoaded = {}

      if self._encrypt:
          fd = open(self._cfgFile, "r")
          encryptedData = fd.read()
          fd.close()
          decryptedString = ConfigManager.Decrypt(encryptedData)
          dictLoaded = json.loads(decryptedString)
      else:

          fp = open(self._cfgFile, 'r')
          dictLoaded = json.load(fp)
          fp.close()

      return dictLoaded

    def load(self, showLoadedMsg=True):
        """@brief Load the config."""
        missingKey= False

        if not isfile(self._cfgFile):

            self._configDict = self._defaultConfig
            self.store()

        else:
            #Get the config from the stored config file.
            loadedConfig = self._getDict()
            #Get list of all the keys from the config file loaded.
            loadedConfigKeys = loadedConfig.keys()
            #Get the default config
            self._configDict = copy.deepcopy( self._defaultConfig )
            #Ensure that the config we use has all the values from the file we just loaded and any other values
            #not in the config file loaded but defined in the default config are set to default values.
            for key in loadedConfigKeys:
                if key in self._configDict:
                    self._configDict[key] = loadedConfig[key]
                    self._uio.debug("{} = {}".format(key, self._configDict[key]))
                else:
                    self._uio.debug("----------> DROPPED FROM CONFIG: {} = {}".format(key, loadedConfig[key]))

        self._uio.info("Loaded config from %s" % (self._cfgFile) )

    def inputFloat(self, key, prompt, minValue=UNSET_VALUE, maxValue=UNSET_VALUE):
        """@brief Input a float value into the config.
           @key The key to store this value in the config.
           @param prompt       The prompt presented to the user in order to enter
                             the float value.
           @param minValue     The minimum acceptable value.
           @param maxValue     The maximum acceptable value."""
        value = ConfigManager.GetFloat(self._uio, prompt, previousValue=self._getValue(key), minValue=minValue, maxValue=maxValue)
        self.addAttr(key, value)

    def inputDecInt(self, key, prompt, minValue=UNSET_VALUE, maxValue=UNSET_VALUE):
        """@brief Input a decimal integer value into the config.
           @key The key to store this value in the config.
           @param prompt       The prompt presented to the user in order to enter
                             the float value.
           @param minValue     The minimum acceptable value.
           @param maxValue     The maximum acceptable value."""
        value = ConfigManager.GetDecInt(self._uio, prompt, previousValue=self._getValue(key), minValue=minValue, maxValue=maxValue)
        self.addAttr(key, value)

    def inputHexInt(self, key, prompt, minValue=UNSET_VALUE, maxValue=UNSET_VALUE):
        """@brief Input a hexadecimal integer value into the config.
           @key The key to store this value in the config.
           @param prompt       The prompt presented to the user in order to enter
                             the float value.
           @param minValue     The minimum acceptable value.
           @param maxValue     The maximum acceptable value."""
        value = ConfigManager.GetHexInt(self._uio, prompt, previousValue=self._getValue(key), minValue=minValue, maxValue=maxValue)
        self.addAttr(key, value)

    def _getValue(self, key):
        """@brief Get the current value of the key.
           @param key The key of the value we're after.
           @return The value of the key or and empty string if key not found."""
        value=""

        if key in self._configDict:
            value = self.getAttr(key)

        return value

    def inputStr(self, key, prompt, allowEmpty):
        """@brief Input a string value into the config.
           @param key The key to store this value in the config.
           @param prompt The prompt presented to the user in order to enter
                         the float value.
           @param allowEmpty   If True then allow the string to be empty."""
        value = ConfigManager.GetString(self._uio, prompt, previousValue=self._getValue(key), allowEmpty=allowEmpty )
        self.addAttr(key, value)

    def inputDate(self, key, prompt, allowEmpty):
        """@brief Input a date into the config.
           @param key The key to store this value in the config.
           @param prompt The prompt presented to the user in order to enter
                         the float value.
           @param allowEmpty   If True then allow the string to be empty."""
        value = ConfigManager.GetDate(self._uio, prompt, previousValue=self._getValue(key), allowEmpty=allowEmpty )
        self.addAttr(key, value)

    def inputTime(self, key, prompt, allowEmpty):
        """@brief Input a date into the config.
           @param key The key to store this value in the config.
           @param prompt The prompt presented to the user in order to enter
                         the float value.
           @param allowEmpty   If True then allow the string to be empty."""
        value = ConfigManager.GetTime(self._uio, prompt, previousValue=self._getValue(key), allowEmpty=allowEmpty )
        self.addAttr(key, value)

    def inputBool(self, key, prompt):
        """@brief Input a boolean value.
           @param key The key to store this value in the config.
           @param prompt The prompt presented to the user in order to enter
                         the boolean (Yes/No) value."""
        previousValue=self._getValue(key)
        yes = self.getYesNo(prompt, previousValue=previousValue)
        if yes:
            value=1
        else:
            value=0
        self.addAttr(key, value)

    def getYesNo(self, prompt, previousValue=0):
        """@brief Input yes no response.
           @param prompt The prompt presented to the user in order to enter
                         the float value.
           @param allowEmpty   If True then allow the string to be empty.
           @return True if Yes, False if No."""

        _prompt = "%s y/n" % (prompt)

        if previousValue:
            prevValue='y'
        else:
            prevValue='n'

        while True:
            value = ConfigManager.GetString(self._uio, _prompt, prevValue)
            value=value.lower()
            if value == 'n':
                return False
            elif value == 'y':
                return True

    def _getModifiedTime(self):
        """@brief Get the modified time of the config file."""
        mtime = 0
        try:

            if isfile(self._cfgFile):
                mtime = getmtime(self._cfgFile)

        except OSError:
            pass

        return mtime

    def updateModifiedTime(self):
        """@brief Update the modified time held as an attr in this nistance with the current modified time of the file."""
        self._modifiedTime = self._getModifiedTime()

    def isModified(self):
        """@Return True if the config file has been updated."""
        mTime = self._getModifiedTime()
        if mTime != self._modifiedTime:
            return True
        return False

    def _getConfigAttDetails(self, key, configAttrDetailsDict):
        """@brief Get the configAttrDetails details instance from the dict.
           @param key The in to the value in the configAttrDetailsDict
           @param configAttrDetailsDict The dict containing attr meta data."""
        if key in configAttrDetailsDict:
            return configAttrDetailsDict[key]
        raise Exception("getConfigAttDetails(): The %s dict has no key=%s" % ( str(configAttrDetailsDict), key) )

    def edit(self, configAttrDetailsDict):
        """@brief A high level method to allow user to edit all config attributes.
           @param configAttrDetailsDict A dict that holds configAttrDetails
                  instances, each of which provide data required for the
                  user to enter the configuration parameter."""

        if len(self._configDict.keys()) == 0:
            self.load(showLoadedMsg=True)

        keyList = list(self._configDict.keys())
        keyList.sort()
        index = 0
        while index < len(keyList):

            try:

                key = keyList[index]

                configAttrDetails = self._getConfigAttDetails(key, configAttrDetailsDict)

                if key.endswith("_FLOAT"):

                    self.inputFloat(key, configAttrDetails.prompt, minValue=configAttrDetails.minValue, maxValue=configAttrDetails.maxValue)

                elif key.endswith("_INT"):

                    self.inputDecInt(key, configAttrDetails.prompt, minValue=configAttrDetails.minValue, maxValue=configAttrDetails.maxValue)

                elif key.endswith("_HEXINT"):

                    self.inputHexInt(key, configAttrDetails.prompt, minValue=configAttrDetails.minValue, maxValue=configAttrDetails.maxValue)

                elif key.endswith("_STR"):

                    self.inputStr(key, configAttrDetails.prompt, configAttrDetails.allowEmpty)

                index = index + 1

            except KeyboardInterrupt:

                if index > 0:
                    index=index-1
                    print('\n')

                else:
                    while True:
                        try:
                            print('\n')
                            if self.getYesNo("Quit ?"):
                                sys.exit(0)
                            break
                        except KeyboardInterrupt:
                            pass

        self.store()

    def show(self, uio):
        """@brief Show the state of configuration parameters.
           @param uio The UIO instance to use to show the parameters."""
        attrList = self.getAttrList()
        attrList.sort()

        maxAttLen=0
        for attr in attrList:
            if len(attr) > maxAttLen:
                maxAttLen=len(attr)

        for attr in attrList:
            padding = " "*(maxAttLen-len(attr))
            uio.info("%s%s = %s" % (attr, padding, self.getAttr(attr)) )

    def getConfigDict(self):
        """@return the dict holding the configuration."""
        return self._configDict

    def setDefaultConfig(self):
        """@brief Set the default configuration by removing the existing configuration file and re loading."""
        configFile = self._getConfigFile()
        if isfile(configFile):
            self._uio.info(configFile)
            deleteFile = self._uio.getBoolInput("Are you sure you wish to delete the above file [y]/[n]")
            if deleteFile:
                os.remove(configFile)
                self._uio.info("{} has been removed.".format(configFile))
                self._uio.info("The default configuration will be loaded next time..")

    def configure(self, editConfigMethod):
        """@brief A helper method to edit the dictionary config.
           @param editConfigMethod The method to call to edit configuration.
           @return None"""
        running=True
        while running:
            idKeyDict=self.show()
            response = self._uio.getInput("Enter 'E' to edit a parameter, 'S' to save and quit or 'Q' to quit")
            response=response.upper()
            if response == 'E':
                id = self._uio.getIntInput("Enter the ID of the parameter to change")
                if id not in idKeyDict:
                    self._uio.error("Configuration ID {} is invalid.".format(id))
                else:
                    key=idKeyDict[id]
                    editConfigMethod(key)
            elif response == 'S':
                self.store()
                running = False
            elif response == 'Q':
                running = False

    def show(self):
        """@brief A helper method to show the dictionary config to the user.
           @return A dictionary mapping the attribute ID's (keys) to dictionary keys (values)."""
        maxIDLen = 3
        maxKeyLen=10
        for key in self._configDict:
            if len(key) > maxKeyLen:
                maxKeyLen = len(key)
        self._uio.info("ID  PARAMETER"+" "*(maxKeyLen-8)+" VALUE")
        id=1
        idKeyDict = {}
        for key in self._configDict:
            idKeyDict[id]=key
            self._uio.info("{:<3d} {:<{}} {}".format(id, key, maxKeyLen+1, self._configDict[key]))
            id=id+1
        return idKeyDict

class ConfigAttrDetails(object):
    """@brief Responsible for holding config attribute meta data."""

    def __init__(self, prompt, minValue=ConfigManager.UNSET_VALUE, maxValue=ConfigManager.UNSET_VALUE, allowEmpty=True):
        self.prompt         =   prompt      #Always used to as k user to enter attribute value.
        self.minValue       =   minValue    #Only used for numbers
        self.maxValue       =   maxValue    #Only used for numbers
        self.allowEmpty     =   allowEmpty  #Only used for strings



#!/usr/bin/env python3

"""This file is responsible for providing general helper functionality not
   associated with particular objects"""

import sys
import os
import platform
import json
import traceback
import  socket

def initArgs(parser, lastCmdLineArg=None, checkHostArg=True):
    """This method is responsible for
        - Ensure that the host argument has been defined by the user on the command line
        - Set the debug level as defined on the command line.
        - If lastCmdLineArg is defined then ensure that this is the last arguemnt on the
          command line. If we don't do this then the user may define some arguments on
          the command line after the action (callback) command line option but as
          the arguments are processed in sequence the args following the action
          arg will not be used.
    """

    if checkHostArg:
        if parser.values.host == None or len(parser.values.host) == 0:
            raise Exception("Please define the RFeye host on the command line.")

    parser.uio.debugLevel = parser.values.debug

    # Check that the argument that invoked tha action (callback) is the last on the command line
    argOk = False
    if len(sys.argv) > 0:
        lastArg = sys.argv[len(sys.argv) - 1]
        if lastArg == lastCmdLineArg:
            argOk = True

    if not argOk:
        raise Exception("Please ensure %s (if used) is the last argument on the command line." % (lastCmdLineArg))


def ensureBoolInt(value, arg):
    """We expect value to be a boolean (0 or 1)
       raise an error if not
    """
    if value not in [0, 1]:
        raise Exception("The %s arg should be followed by 0 or 1. Was %s." % (str(arg), str(value)))


def getLines(text):
    """Split the text into lines"""
    lines = []
    if len(text) > 0:
        elems = text.split("\n")
        lines = stripEOL(elems)
    return lines


def stripEOL(lines):
    """Strip the end of line characters from the list of lines of text"""
    noEOLLines = []
    for l in lines:
        l = l.rstrip("\n")
        l = l.rstrip("\r")
        noEOLLines.append(l)
    return noEOLLines


def getLinesFromFile(f):
    """Get Lines from file"""
    fd = open(f, "r")
    lines = fd.readlines()
    fd.close()
    return stripEOL(lines)


def _removeInvalidChars(line):
    """Return a copy of line with each ASCII control character (0-31),
    and each double quote, removed."""
    output = ''
    for c in line:
        if c >= ' ' and c != '"':
            output = output + c
    return output


def _addEntry(line, dict):
    """Parse line into a key and value, adding the result to dict, as in
    getDict."""
    # check for a parameter
    fields = line.split('=')
    # if at least 2 fields exist
    if len(fields) > 1:
        # add the key,value pair to the dictionary
        key = _removeInvalidChars(fields[0])
        value = _removeInvalidChars(fields[1])
        dict[key] = value


def getDict(filename, jsonFmt=False):
    """@brief Load dict from file
       @param jsonFmt If True then we expect the file to be in json format.

           if json is True we expect the file to be in json format

           if json is False
            We key=value pairs (= is the separate character).
            Lines containing a hash sign as the first non-whitespace character are
            ignored. Leading and trailing whitespace is ignored.

            Lines not containing an equals sign are also silently ignored.

            Lines not ignored are assumed to be in the form key=value, where key
            does not contain an equals sign;
            Control characters and double quotes in both key and value are silently
            discarded. value is also truncated just before the first whitespace or
            equals sign it contains.

       @return Return the dict loaded from the file.

    """
    dictLoaded = {}

    if jsonFmt:
        fp = open(filename, 'r')
        dictLoaded = json.load(fp)
        fp.close()

    else:

        lines = getLinesFromFile(filename)
        for line in lines:
            # strip leading and trailing whitespaces
            line = line.strip()
            # if a comment line then ignore
            if line.find('#') == 0:
                continue
            # add an entry to the dict
            _addEntry(line, dictLoaded)

    return dictLoaded


def saveDict(dictToSave, filename, jsonFmt=False):
    """@brief Save dict to a file.
       @param jsonFmt If True then we expect the file to be in json format.

       if json is True we expect the file to be in json format

       if json is False the file is saved as key = value pairs
       Each key in dict produces a line of the form key=value. Output will be
       ambiguous if any keys contain equals signs or if any values contain
       newlines.

    """

    if jsonFmt:
        try:

            with open(filename, "w") as write_file:
                json.dump(dictToSave, write_file)

        except IOError as i:
            raise IOError(i.errno, 'Failed to write file \'%s\': %s'
                          % (filename, i.strerror), i.filename).with_traceback(sys.exc_info()[2])
    else:
        lines = []
        # build config file lines
        for key in list(dictToSave.keys()):
            lines.append(str(key) + '=' + str(dictToSave.get(key)) + '\n')
        try:
            f = open(filename, 'w')
            f.writelines(lines)
            f.close()
        except IOError as i:
            raise IOError(i.errno, 'Failed to write file \'%s\': %s'
                          % (filename, i.strerror), i.filename).with_traceback(sys.exc_info()[2])


def getAddrPort(host):
    """The host address may be entered in the format <address>:<port>
       Return a tuple with host and port"""
    elems = host.split(":")

    if len(elems) > 1:
        host = elems[0]
        port = int(elems[1])
    else:
        port = 22

    return [host, port]


def getProgramName():
    """Get the name of the currently running program."""
    progName = sys.argv[0].strip()
    if progName.startswith('./'):
        progName = progName[2:]
    if progName.endswith('.py'):
        progName = progName[:-3]

    # Only return the name of the program not the path
    pName = os.path.split(progName)[-1]
    if pName.endswith('.exe'):
        pName = pName[:-4]
    return pName


def getBoolUserResponse(uio, prompt, allowQuit=True):
    """Get boolean (Y/N) repsonse from user.
       If allowQuit is True and the user enters q then the program will exit."""
    while True:
        response = uio.getInput(prompt=prompt)
        if response.lower() == 'y':
            return True
        elif response.lower() == 'n':
            return False
        elif allowQuit and response.lower() == 'q':
            sys.exit(0)


def getIntUserResponse(uio, prompt, allowQuit=True):
    """Get int repsonse from user.
       If allowQuit is True and the user enters q then None is returned to
       indicate that the user selected quit."""
    while True:
        response = uio.getInput(prompt=prompt)

        try:

            return int(response)

        except ValueError:

            uio.info("%s is not a valid integer value." % (response))

        if allowQuit and response.lower() == 'q':
            return None


def getIntListUserResponse(uio, prompt, minValue=None, maxValue=None, allowQuit=True):
    """Get int repsonse from user as a list of int's
       If allowQuit is True and the user enters q then the program will exit."""
    while True:
        response = uio.getInput(prompt=prompt)
        try:
            elems = response.split(",")
            if len(elems) > 0:
                intList = []
                errorStr = None
                for vStr in elems:
                    v = int(vStr)
                    if minValue != None and v < minValue:
                        errorStr = "The min value that may be entered is %d." % (minValue)
                        break
                    elif maxValue != None and v > maxValue:
                        errorStr = "The max value that may be entered is %d." % (maxValue)
                        break
                    else:
                        intList.append(v)

                if errorStr != None:
                    uio.error(errorStr)

            return intList
        except ValueError:
            pass
        if allowQuit and response.lower() == 'q':
            sys.exit(0)


def getHomePath():
    """Get the user home path as this will be used to store config files"""
    if "HOME" in os.environ:
        return os.environ["HOME"]

    elif "HOMEDRIVE" in os.environ and "HOMEPATH" in os.environ:
        return os.environ["HOMEDRIVE"] + os.environ["HOMEPATH"]

    elif "USERPROFILE" in os.environ:
        return os.environ["USERPROFILE"]

    return None


def setHomePath(homePath):
    """Seth the env variable HOME"""
    # Do some sanity/defensive stuff
    if homePath == None:
        raise Exception("homePath=None.")
    elif len(homePath) == 0:
        raise Exception("len(homePath)=0.")

    # We do some special stuff on windows
    if platform.system() == "Windows":

        os.environ["HOME"] = homePath

        if "USERPROFILE" in os.environ:
            os.environ["USERPROFILE"] = os.environ["HOME"]

        if "HOMEPATH" in os.environ:
            os.environ["HOMEPATH"] = os.environ["HOME"]

        if not os.path.isdir(os.environ["HOME"]):
            raise Exception(os.environ["HOME"] + " path not found.")

    else:
        # Not windows set HOME env var
        os.environ["HOME"] = homePath

        if not os.path.isdir(os.environ["HOME"]):
            raise Exception(os.environ["HOME"] + " path not found.")

def printDict(uio, theDict, indent=0):
    """@brief Show the details of a dictionary contents
       @param theDict The dictionary
       @param indent Number of tab indents
       @return None"""
    for key in theDict:
        uio.info('\t' * indent + str(key))
        value = theDict[key]
        if isinstance(value, dict):
            printDict(uio, value, indent + 1)
        else:
            uio.info('\t' * (indent + 1) + str(value))

def logTraceBack(uio):
    """@brief Log a traceback using the uio instance to the debug file.
       @param uio A UIO instance
       @return None"""
    # Always store the exception traceback in the logfile as this makes
    # it easier to diagnose problems with the testing
    lines = traceback.format_exc().split("\n")
    for line in lines:
        uio.storeToDebugLog(line)

def GetFreeTCPPort():
    """@brief Get a free port and return to the client. If no port is available
              then -1 is returned.
       @return the free TCP port number or -1 if no port is available."""
    tcpPort=-1
    try:
        #Bind to a local port to find a free TTCP port
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('', 0))
        tcpPort = sock.getsockname()[1]
        sock.close()
    except socket.error:
        pass
    return tcpPort

def appendCreateFile(uio, aFile, quiet=False):
    """@brief USer interaction to append or create a file.
       @param uio A UIO instance.
       @param quiet If True do not show uio messages (apart from overwrite prompt.
       @param aFile The file to append or delete."""
    createFile = False
    if os.path.isfile(aFile):
        if uio.getBoolInput("Overwrite {} y/n".format(aFile)):
            os.remove(aFile)
            if not quiet:
                uio.info("Deleted {}".format(aFile))
            createFile = True
        else:
            if not quiet:
                uio.info("Appending to {}".format(aFile))

    else:
        createFile = True

    if createFile:
        fd = open(aFile, 'w')
        fd.close()
        if not quiet:
            uio.info("Created {}".format(aFile))
